from vega_sim.environment.agent import Agent, StateAgentWithWallet
from vega_sim.scenario.common.agents import MarketHistoryData, ResourceData
from typing import List, Callable, Optional, Dict
import pandas as pd
import itertools
import os
import os.path
import json

import vega_sim.proto.vega as vega_protos

DEFAULT_PATH = "./run_logs"
DEFAULT_RUN_NAME = "latest"

AGENTS_FILE_NAME = "agents.csv"
DATA_FILE_NAME = "market_data.csv"
ORDER_BOOK_FILE_NAME = "depth_data.csv"
TRADES_FILE_NAME = "trades.csv"
ACCOUNTS_FILE_NAME = "accounts.csv"
FUZZING_FILE_NAME = "additional_data.csv"
RESOURCES_FILE_NAME = "resources.csv"
ASSETS_FILE_NAME = "assets.csv"
MARKET_CHAIN_FILE_NAME = "market_chain.json"


def resource_data_to_row(data: ResourceData):
    return [
        {
            "time": data.at_time,
            "vega_cpu_per": data.vega_cpu_per,
            "vega_mem_rss": data.vega_mem_rss,
            "vega_mem_vms": data.vega_mem_vms,
            "datanode_cpu_per": data.datanode_cpu_per,
            "datanode_mem_rss": data.datanode_mem_rss,
            "datanode_mem_vms": data.datanode_mem_vms,
        }
    ]


def history_data_to_row(data: MarketHistoryData) -> List[pd.Series]:
    results = []
    for market_id in data.market_info.keys():
        market_data = data.market_data[market_id]

        results.append(
            {
                "time": data.at_time,
                "mark_price": market_data.mark_price,
                "market_id": market_id,
                "mark_price": market_data.mark_price,
                "mid_price": market_data.mid_price,
                "open_interest": market_data.open_interest,
                "best_bid": market_data.best_bid_price,
                "best_offer": market_data.best_offer_price,
                "best_bid_volume": market_data.best_bid_volume,
                "best_offer_volume": market_data.best_offer_volume,
                "market_state": market_data.market_state,
                "market_trading_mode": market_data.market_trading_mode,
                "target_stake": market_data.target_stake,
                "supplied_stake": market_data.supplied_stake,
                "price_monitoring_bounds": [
                    ((bound.min_valid_price, bound.max_valid_price))
                    for bound in market_data.price_monitoring_bounds
                ],
                "indicative_price": market_data.indicative_price,
                "trigger": market_data.trigger,
                "extension_trigger": market_data.extension_trigger,
            },
        )
    return results


def history_data_to_order_book_rows(data: MarketHistoryData) -> List[pd.Series]:
    results = []
    for market_id, depth in data.market_depth.items():
        for side, side_vals in [("BID", depth.buys), ("ASK", depth.sells)]:
            for i, level_data in enumerate(side_vals):
                results.append(
                    {
                        "time": data.at_time,
                        "side": side,
                        "price": level_data.price,
                        "volume": level_data.volume,
                        "level": i,
                        "market_id": market_id,
                    }
                )
    return results


def history_data_to_trade_rows(data: MarketHistoryData) -> List[pd.Series]:
    results = []
    for market_id, trades in data.trades.items():
        for trade in trades:
            results.append(
                {
                    "time": trade.timestamp,
                    "seen_at": data.at_time,
                    "id": trade.id,
                    "price": trade.price,
                    "size": trade.size,
                    "buyer": trade.buyer,
                    "seller": trade.seller,
                    "aggressor": trade.aggressor,
                    "buy_order": trade.buy_order,
                    "sell_order": trade.sell_order,
                    "market_id": market_id,
                    "trade_type": trade.trade_type,
                    "buyer_fee_infrastructure": trade.buyer_fee.infrastructure_fee,
                    "buyer_fee_liquidity": trade.buyer_fee.liquidity_fee,
                    "buyer_fee_maker": trade.buyer_fee.maker_fee,
                    "seller_fee_infrastructure": trade.seller_fee.infrastructure_fee,
                    "seller_fee_liquidity": trade.seller_fee.liquidity_fee,
                    "seller_fee_maker": trade.seller_fee.maker_fee,
                    "buyer_auction_batch": trade.buyer_auction_batch,
                    "seller_auction_batch": trade.seller_auction_batch,
                }
            )
    return results


def history_data_to_account_rows(data: MarketHistoryData) -> List[pd.Series]:
    results = []
    for account in data.accounts:
        results.append(
            {
                "time": data.at_time,
                "party_id": account.owner,
                "balance": account.balance,
                "market_id": account.market_id,
                "asset": account.asset,
                "type": account.type,
            }
        )
    return results


def _market_data_standard_output(
    market_history_data: List[MarketHistoryData],
    file_name: str,
    output_path: str = DEFAULT_PATH,
    data_to_row_fn: Callable[[MarketHistoryData], pd.Series] = history_data_to_row,
):
    results = list(
        itertools.chain.from_iterable(
            [data_to_row_fn(data=step_data) for step_data in market_history_data],
        )
    )
    result_df = pd.DataFrame.from_records(
        results,
        index="time" if len(results) > 0 else None,
    )
    os.makedirs(output_path, exist_ok=True)

    result_df.to_csv(path_or_buf=os.path.join(output_path, file_name))


def market_data_standard_output(
    market_history_data: List[MarketHistoryData],
    run_name: str = DEFAULT_RUN_NAME,
    output_path: str = DEFAULT_PATH,
    custom_output_fns: Optional[
        Dict[str, List[Callable[[MarketHistoryData], pd.Series]]]
    ] = None,
):
    run_name = run_name if run_name is not None else DEFAULT_RUN_NAME
    full_path = os.path.join(output_path, run_name)
    data_fns = (
        custom_output_fns
        if custom_output_fns is not None
        else {
            DATA_FILE_NAME: history_data_to_row,
            ORDER_BOOK_FILE_NAME: history_data_to_order_book_rows,
            TRADES_FILE_NAME: history_data_to_trade_rows,
            ACCOUNTS_FILE_NAME: history_data_to_account_rows,
        }
    )
    for file_name, data_fn in data_fns.items():
        _market_data_standard_output(
            market_history_data=market_history_data,
            file_name=file_name,
            output_path=full_path,
            data_to_row_fn=data_fn,
        )


def resources_standard_output(
    resource_data: List[ResourceData],
    run_name: str = DEFAULT_RUN_NAME,
    output_path: str = DEFAULT_PATH,
    custom_output_fns: Optional[
        Dict[str, List[Callable[[MarketHistoryData], pd.Series]]]
    ] = None,
):
    run_name = run_name if run_name is not None else DEFAULT_RUN_NAME
    full_path = os.path.join(output_path, run_name)
    data_fns = (
        custom_output_fns
        if custom_output_fns is not None
        else {
            RESOURCES_FILE_NAME: resource_data_to_row,
        }
    )
    for file_name, data_fn in data_fns.items():
        _market_data_standard_output(
            market_history_data=resource_data,
            file_name=file_name,
            output_path=full_path,
            data_to_row_fn=data_fn,
        )


def agents_standard_output(
    agents: Dict[str, Agent],
    run_name: str = DEFAULT_RUN_NAME,
    output_path: str = DEFAULT_PATH,
):
    results = {
        key: [] for key in ["agent_name", "agent_type", "agent_tag", "agent_key"]
    }
    for name, agent in agents.items():
        results["agent_name"].append(name)
        results["agent_tag"].append(agent.tag)
        results["agent_type"].append(type(agent).__name__)
        if isinstance(agent, StateAgentWithWallet):
            results["agent_key"].append(agent._public_key)
        else:
            results["agent_key"].append(None)

    full_path = os.path.join(output_path, run_name)
    os.makedirs(full_path, exist_ok=True)
    pd.DataFrame.from_dict(results).set_index(
        "agent_name",
        drop=True,
    ).to_csv(os.path.join(full_path, AGENTS_FILE_NAME))


def assets_standard_output(
    assets: List[vega_protos.assets.Asset],
    run_name: str = DEFAULT_RUN_NAME,
    output_path: str = DEFAULT_PATH,
):
    results = {
        key: []
        for key in [
            "id",
            "name",
            "symbol",
            "decimals",
            "quantum",
            "max_faucet_amount_mint",
            "status",
        ]
    }
    for asset in assets:
        results["id"].append(asset.id)
        results["name"].append(asset.details.name)
        results["symbol"].append(asset.details.symbol)
        results["decimals"].append(asset.details.decimals)
        results["quantum"].append(asset.details.quantum)
        results["max_faucet_amount_mint"].append(
            asset.details.builtin_asset.max_faucet_amount_mint
        )
        results["status"].append(asset.status)

    full_path = os.path.join(output_path, run_name)
    os.makedirs(full_path, exist_ok=True)
    pd.DataFrame.from_dict(results).set_index("id", drop=True).to_csv(
        os.path.join(full_path, ASSETS_FILE_NAME)
    )


def market_chain_standard_output(
    market_history_data: List[MarketHistoryData],
    run_name: str = DEFAULT_RUN_NAME,
    output_path: str = DEFAULT_PATH,
):
    run_name = run_name if run_name is not None else DEFAULT_RUN_NAME
    full_path = os.path.join(output_path, run_name)

    market_chains: Dict[str, List[str]] = {}
    for history_data in market_history_data:
        for market_id, market_info in history_data.market_info.items():
            if not market_info.parent_market_id:
                market_chains.setdefault(market_id, [market_id])
            else:
                for _, market_chain in market_chains.items():
                    if market_chain[-1] == market_info.parent_market_id:
                        market_chain.append(market_id)

    os.makedirs(output_path, exist_ok=True)

    with open(os.path.join(full_path, MARKET_CHAIN_FILE_NAME), "w") as f:
        json.dump(market_chains, f, indent=4)


def load_agents_df(
    run_name: Optional[str] = None,
    output_path: str = DEFAULT_PATH,
) -> pd.DataFrame:
    run_name = run_name if run_name is not None else DEFAULT_RUN_NAME
    return pd.read_csv(
        os.path.join(output_path, run_name, AGENTS_FILE_NAME), index_col="agent_name"
    )


def load_assets_df(
    run_name: Optional[str] = None,
    output_path: str = DEFAULT_PATH,
) -> pd.DataFrame:
    run_name = run_name if run_name is not None else DEFAULT_RUN_NAME
    return pd.read_csv(
        os.path.join(output_path, run_name, ASSETS_FILE_NAME), index_col="id"
    )


def load_market_data_df(
    run_name: Optional[str] = None,
    output_path: str = DEFAULT_PATH,
) -> pd.DataFrame:
    run_name = run_name if run_name is not None else DEFAULT_RUN_NAME
    df = pd.read_csv(os.path.join(output_path, run_name, DATA_FILE_NAME))
    if not df.empty:
        df["time"] = pd.to_datetime(df.time)
        df = df.set_index("time")
    return df


def load_order_book_df(
    run_name: Optional[str] = None,
    output_path: str = DEFAULT_PATH,
) -> pd.DataFrame:
    run_name = run_name if run_name is not None else DEFAULT_RUN_NAME
    depth_df = pd.read_csv(os.path.join(output_path, run_name, ORDER_BOOK_FILE_NAME))
    if not depth_df.empty:
        depth_df["time"] = pd.to_datetime(depth_df.time)
        depth_df = depth_df[depth_df["time"] != depth_df["time"].min()].set_index(
            "time"
        )
    return depth_df


def load_trades_df(
    run_name: Optional[str] = None,
    output_path: str = DEFAULT_PATH,
) -> pd.DataFrame:
    run_name = run_name if run_name is not None else DEFAULT_RUN_NAME
    df = pd.read_csv(os.path.join(output_path, run_name, TRADES_FILE_NAME))
    if not df.empty:
        df["time"] = pd.to_datetime(df.time)
        df = df.set_index("time")
    return df


def load_accounts_df(
    run_name: Optional[str] = None,
    output_path: str = DEFAULT_PATH,
) -> pd.DataFrame:
    run_name = run_name if run_name is not None else DEFAULT_RUN_NAME
    df = pd.read_csv(os.path.join(output_path, run_name, ACCOUNTS_FILE_NAME))
    if not df.empty:
        df["time"] = pd.to_datetime(df.time)
        df = df.set_index("time")
    return df


def load_fuzzing_df(
    run_name: Optional[str] = None,
    output_path: str = DEFAULT_PATH,
) -> pd.DataFrame:
    run_name = run_name if run_name is not None else DEFAULT_RUN_NAME
    df = pd.read_csv(os.path.join(output_path, run_name, FUZZING_FILE_NAME))
    if not df.empty:
        df["time"] = pd.to_datetime(df.time)
        df = df.set_index("time")
    return df


def load_resource_df(
    run_name: Optional[str] = None,
    output_path: str = DEFAULT_PATH,
) -> pd.DataFrame:
    run_name = run_name if run_name is not None else DEFAULT_RUN_NAME
    df = pd.read_csv(os.path.join(output_path, run_name, RESOURCES_FILE_NAME))
    if not df.empty:
        df["time"] = pd.to_datetime(df.time)
        df = df.set_index("time")
    return df


def load_market_chain(
    run_name: Optional[str] = None,
    output_path: str = DEFAULT_PATH,
) -> Dict[str, List[str]]:
    run_name = run_name if run_name is not None else DEFAULT_RUN_NAME
    file_path = os.path.join(output_path, run_name, MARKET_CHAIN_FILE_NAME)
    if os.path.exists(file_path):
        with open(file_path, "r") as f:
            market_chain = json.load(fp=f)
    else:
        market_chain = {}
    return market_chain



from __future__ import annotations

import copy
import grpc
import logging
import threading
import traceback
from collections import defaultdict
from itertools import chain, product
from queue import Empty, Queue
from types import GeneratorType
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union

import vega_sim.api.data as data
import vega_sim.api.data_raw as data_raw
import vega_sim.api.governance as gov
import vega_sim.grpc.client as vac
import vega_sim.proto.vega as vega_protos
import vega_sim.proto.vega.events.v1.events_pb2 as events_protos
from vega_sim.tools.retry import retry

logger = logging.getLogger(__name__)


def _queue_forwarder(
    data_client: vac.VegaCoreClient,
    stream_registry: List[
        Tuple[
            Any,
            Callable[
                [
                    vega_protos.api.v1.core.ObserveEventBusResponse,
                    vac.VegaTradingDataClientV2,
                ],
                Any,
            ],
        ]
    ],
    sink: Queue[Any],
    market_id: Optional[str] = None,
    party_id: Optional[str] = None,
    kill_thread_sig: Optional[threading.Event()] = None,
) -> None:
    obs = data_raw.observe_event_bus(
        data_client=data_client,
        type=list(chain(*[ev[0] for ev in stream_registry])),
        market_id=market_id,
        party_id=party_id,
    )

    handlers = {}
    for evts, handler in stream_registry:
        for evt in evts:
            handlers[evt] = handler
    try:
        for o in obs:
            for event in o.events:
                if (kill_thread_sig is not None) and kill_thread_sig.is_set():
                    return
                output = retry(5, 1.0, lambda: handlers[event.type](event))
                if isinstance(output, (list, GeneratorType)):
                    for elem in output:
                        sink.put(elem)
                else:
                    sink.put(output)
    except grpc._channel._MultiThreadedRendezvous as e:
        if e.details() in ["Channel closed!", "Socket closed"]:
            logging.debug(f"Thread finished as {e.details}")
            logger.info("Data cache event bus closed.")
        else:
            raise e


class DecimalsCache(defaultdict):
    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError(key)
        else:
            ret = self[key] = self.default_factory(key)
            return ret


class LocalDataCache:
    def __init__(
        self,
        event_bus_client: Union[vac.VegaTradingDataClientV2, vac.VegaCoreClient],
        trading_data_client: vac.VegaTradingDataClientV2,
        market_pos_decimals: Optional[Dict[str, int]] = None,
        market_price_decimals: Optional[Dict[str, int]] = None,
        asset_decimals: Optional[Dict[str, int]] = None,
        market_to_asset: Optional[Dict[str, str]] = None,
    ):
        """ """
        self._trading_data_client = trading_data_client
        self._event_bus_client = event_bus_client

        self._market_price_decimals = market_price_decimals
        self._market_pos_decimals = market_pos_decimals
        self._asset_decimals = asset_decimals
        self._market_to_asset = market_to_asset

        self.time_update_lock = threading.RLock()
        self.orders_lock = threading.RLock()
        self.transfers_lock = threading.RLock()
        self.asset_lock = threading.RLock()
        self.market_lock = threading.RLock()
        self.market_data_lock = threading.RLock()
        self.trades_lock = threading.RLock()
        self.ledger_entries_lock = threading.RLock()
        self._time_update_from_feed = 0
        self._live_order_state_from_feed = {}
        self._dead_order_state_from_feed = {}
        self._asset_from_feed = {}
        self._market_from_feed = {}
        self.market_data_from_feed_store = {}
        self._transfer_state_from_feed = {}
        self._trades_from_feed: List[data.Trade] = []
        self._ledger_entries_from_feed: List[data.LedgerEntry] = []

        self._observation_thread = None
        self._aggregated_observation_feed: Queue[Any] = Queue()
        self._kill_thread_sig = threading.Event()

        self.stream_registry = [
            (
                (events_protos.BUS_EVENT_TYPE_TIME_UPDATE,),
                lambda evt: evt.time_update,
            ),
            (
                (events_protos.BUS_EVENT_TYPE_TRADE,),
                lambda evt: data.trades_subscription_handler(
                    evt,
                    self._market_pos_decimals,
                    self._market_price_decimals,
                    self._market_to_asset,
                    self._asset_decimals,
                ),
            ),
            (
                (events_protos.BUS_EVENT_TYPE_ASSET,),
                lambda evt: evt.asset,
            ),
            (
                (events_protos.BUS_EVENT_TYPE_MARKET_CREATED,),
                lambda evt: evt.market_created,
            ),
            (
                (events_protos.BUS_EVENT_TYPE_ORDER,),
                lambda evt: data.order_subscription_handler(
                    evt,
                    self._market_pos_decimals,
                    self._market_price_decimals,
                    self._market_to_asset,
                    self._asset_decimals,
                ),
            ),
            (
                (events_protos.BUS_EVENT_TYPE_MARKET_DATA,),
                lambda evt: data.market_data_subscription_handler(
                    evt,
                    self._market_pos_decimals,
                    self._market_price_decimals,
                    self._market_to_asset,
                    self._asset_decimals,
                ),
            ),
        ]
        self._high_load_stream_registry = [
            (
                (events_protos.BUS_EVENT_TYPE_LEDGER_MOVEMENTS,),
                lambda evt: data.ledger_entries_subscription_handler(
                    evt, self._asset_decimals
                ),
            ),
            (
                (events_protos.BUS_EVENT_TYPE_TRANSFER,),
                lambda evt: data.transfer_subscription_handler(
                    evt,
                    self._market_pos_decimals,
                    self._market_price_decimals,
                    self._market_to_asset,
                    self._asset_decimals,
                ),
            ),
        ]

    def stop(self) -> None:
        self._kill_thread_sig.set()
        self._observation_thread.join()
        self._forwarding_thread.join()

    def time_update_from_feed(
        self,
    ) -> int:
        return self._time_update_from_feed

    def asset_from_feed(
        self,
        asset_id: str,
    ) -> vega_protos.assets.Asset:
        if asset_id not in self._asset_from_feed:
            self.initialise_assets()
        return self._asset_from_feed[asset_id]

    def market_from_feed(
        self,
        market_id: str,
    ) -> vega_protos.markets.Market:
        if market_id not in self._market_from_feed:
            self.initialise_markets()
        return self._market_from_feed[market_id]

    def market_data_from_feed(
        self,
        market_id: str,
    ) -> vega_protos.vega.MarketData:
        """
        Output market info.
        """
        return self.market_data_from_feed_store[market_id]

    def order_status_from_feed(
        self, live_only: bool = True
    ) -> Dict[str, Dict[str, Dict[str, data.Order]]]:
        """Returns a copy of current order status based on Order feed if started.
        If order feed has not been started, dict will be empty

        Args:
            live_only:
                bool, default True, whether to filter out dead/cancelled deals
                    from result

        Returns:
            Dictionary mapping market ID -> Party ID -> Order ID -> Order detaails"""
        with self.orders_lock:
            order_dict = copy.copy(self._live_order_state_from_feed)
            if not live_only:
                order_dict.update(self._dead_order_state_from_feed)
        return order_dict

    def orders_for_party_from_feed(
        self,
        party_id: str,
        market_id: str,
        live_only: bool = True,
    ) -> Dict[str, data.Order]:
        return (
            self.order_status_from_feed(live_only=live_only)
            .get(market_id, {})
            .get(party_id, {})
        )

    def transfer_status_from_feed(
        self, live_only: bool = True, blockchain_time: Optional[int] = None
    ):
        with self.transfers_lock:
            transfers_dict = {}
            for party_id, party_transfers in self._transfer_state_from_feed.items():
                for transfer_id, transfer in party_transfers.items():
                    deliver_on = int(transfer.one_off.deliver_on)
                    if not live_only or (
                        deliver_on != 0 and blockchain_time < deliver_on
                    ):
                        transfers_dict.setdefault(party_id, {})[transfer_id] = transfer
        return transfers_dict

    def start_live_feeds(
        self,
        market_ids: Optional[Union[str, List[str]]] = None,
        party_ids: Optional[Union[str, List[str]]] = None,
        start_high_load_feeds: bool = False,
    ):
        market_ids = (
            (market_ids if isinstance(market_ids, list) else [market_ids])
            if market_ids is not None
            else None
        )
        party_ids = (
            (party_ids if isinstance(party_ids, list) else [party_ids])
            if party_ids is not None
            else None
        )
        self.initialise_assets()
        self.initialise_markets()
        self.initialise_time_update_monitoring()
        self.initialise_order_monitoring(
            market_ids=market_ids,
            party_ids=party_ids,
        )
        self.initialise_transfer_monitoring()
        self.initialise_market_data(
            market_ids,
        )

        self._observation_thread = threading.Thread(
            target=self._monitor_stream, daemon=True
        )
        self._observation_thread.start()

        self._forwarding_thread = threading.Thread(
            target=_queue_forwarder,
            args=(
                self._event_bus_client,
                self.stream_registry
                + (self._high_load_stream_registry if start_high_load_feeds else []),
                self._aggregated_observation_feed,
                (
                    (market_ids[0] if len(market_ids) == 1 else None)
                    if market_ids is not None
                    else None
                ),
                (
                    (party_ids[0] if len(party_ids) == 1 else None)
                    if party_ids is not None
                    else None
                ),
                self._kill_thread_sig,
            ),
            daemon=True,
        )
        self._forwarding_thread.start()

    def initialise_assets(self):
        base_assets = data_raw.list_assets(data_client=self._trading_data_client)

        with self.asset_lock:
            for asset in base_assets:
                self._asset_from_feed[asset.id] = asset

    def initialise_markets(self):
        base_markets = data_raw.all_markets(self._trading_data_client)

        with self.market_lock:
            for market in base_markets:
                self._market_from_feed[market.id] = market

    def initialise_time_update_monitoring(self):
        base_time_update = gov.get_blockchain_time(
            data_client=self._trading_data_client
        )
        with self.time_update_lock:
            self._time_update_from_feed = base_time_update

    def initialise_order_monitoring(
        self,
        market_ids: Optional[List[str]] = None,
        party_ids: Optional[List[str]] = None,
    ):
        base_orders = []
        for market_party_tuple in list(
            product(
                (market_ids if market_ids is not None else [None]),
                (party_ids if party_ids is not None else [None]),
            )
        ):
            base_orders.extend(
                data.list_orders(
                    data_client=self._trading_data_client,
                    market_id=market_party_tuple[0],
                    party_id=market_party_tuple[1],
                    live_only=True,
                    price_decimals=(
                        self._market_price_decimals[market_party_tuple[0]]
                        if market_party_tuple[0] is not None
                        else None
                    ),
                    position_decimals=(
                        self._market_pos_decimals[market_party_tuple[0]]
                        if market_party_tuple[0] is not None
                        else None
                    ),
                )
            )

        with self.orders_lock:
            for order in base_orders:
                self._live_order_state_from_feed.setdefault(
                    order.market_id, {}
                ).setdefault(order.party_id, {})[order.id] = order

    def initialise_market_data(
        self,
        market_ids: Optional[List[str]] = None,
    ):
        if market_ids is None:
            market_ids = [
                market.id for market in data_raw.all_markets(self._trading_data_client)
            ]
        with self.market_data_lock:
            for market_id in market_ids:
                self.market_data_from_feed_store[
                    market_id
                ] = data.get_latest_market_data(
                    market_id,
                    data_client=self._trading_data_client,
                    market_price_decimals_map=self._market_price_decimals,
                    market_position_decimals_map=self._market_pos_decimals,
                    asset_decimals_map=self._asset_decimals,
                    market_to_asset_map=self._market_to_asset,
                )

    def initialise_transfer_monitoring(
        self,
    ):
        base_transfers = []

        base_transfers.extend(
            data.list_transfers(data_client=self._trading_data_client)
        )

        with self.transfers_lock:
            for t in base_transfers:
                self._transfer_state_from_feed.setdefault(t.party_to, {})[t.id] = t

    def _monitor_stream(self) -> None:
        while not self._kill_thread_sig.is_set():
            try:
                update = self._aggregated_observation_feed.get(timeout=1)
            except Empty:
                continue
            else:
                if isinstance(update, data.Order):
                    with self.orders_lock:
                        if update.version >= getattr(
                            self._live_order_state_from_feed.setdefault(
                                update.market_id, {}
                            )
                            .setdefault(update.party_id, {})
                            .get(update.id, None),
                            "version",
                            0,
                        ):
                            if (
                                not update.status
                                == vega_protos.vega.Order.Status.STATUS_ACTIVE
                            ):
                                # If the order is dead, pop any we've seen from
                                # live state
                                self._live_order_state_from_feed[update.market_id][
                                    update.party_id
                                ].pop(update.id, None)

                                # And add to dead instead
                                self._dead_order_state_from_feed.setdefault(
                                    update.market_id, {}
                                ).setdefault(update.party_id, {})[update.id] = update
                            else:
                                self._live_order_state_from_feed[update.market_id][
                                    update.party_id
                                ][update.id] = update

                elif isinstance(update, data.Transfer):
                    with self.transfers_lock:
                        self._transfer_state_from_feed.setdefault(update.party_to, {})[
                            update.id
                        ] = update

                elif isinstance(update, data.Trade):
                    with self.trades_lock:
                        self._trades_from_feed.append(update)

                elif isinstance(update, data.MarketData):
                    with self.market_data_lock:
                        self.market_data_from_feed_store[update.market_id] = update

                elif isinstance(update, data.LedgerEntry):
                    with self.ledger_entries_lock:
                        self._ledger_entries_from_feed.append(update)

                elif isinstance(update, events_protos.TimeUpdate):
                    with self.time_update_lock:
                        self._time_update_from_feed = int(update.timestamp)

                elif isinstance(update, vega_protos.markets.Market):
                    self._market_from_feed[update.id] = update
                    # Trigger the DecimalCache default factory method by attempting to
                    # get the decimal precision for the market
                    self._market_pos_decimals[update.id]
                    self._market_price_decimals[update.id]
                    self._market_to_asset[update.id]

                elif isinstance(update, vega_protos.assets.Asset):
                    self._asset_from_feed[update.id] = update
                    # Trigger the DecimalCache default factory method by attempting to
                    # get the decimal precision for the asset
                    self._asset_decimals[update.id]

                elif update is None:
                    logger.debug("Failed to process event into update.")

                else:
                    logger.info(f"Unhandled update {update}")

    def get_ledger_entries_from_stream(
        self,
        party_id_from: Optional[str] = None,
        party_id_to: Optional[str] = None,
        transfer_type: Optional[str] = None,
    ) -> List[data.LedgerEntry]:
        results = []

        for ledger_entry in self._ledger_entries_from_feed:
            if (
                transfer_type is not None
                and transfer_type != ledger_entry.transfer_type
            ):
                continue
            if (
                party_id_from is not None
                and party_id_from != ledger_entry.from_account.owner
            ):
                continue
            if party_id_to is not None and party_id_to != ledger_entry.to_account.owner:
                continue

            results.append(copy.copy(ledger_entry))

        return results

    def get_trades_from_stream(
        self,
        market_id: Optional[str] = None,
        party_id: Optional[str] = None,
        order_id: Optional[str] = None,
        exclude_trade_ids: Optional[Set[str]] = None,
    ) -> List[data.Trade]:
        """Loads executed trades for a given query of party/market/specific order from
        data node. Converts values to proper decimal output.

        Args:
            market_id:
                optional str, Restrict to trades on a specific market
            party_id:
                optional str, Select only trades with a given id
            order_id:
                optional str, Restrict to trades for a specific order
            exclude_trade_ids:
                optional set[str], Do not return trades with ID in this set

        Returns:
            List[Trade], list of formatted trade objects which match the required
                restrictions.
        """
        with self.trades_lock:
            results = []
            for trade in self._trades_from_feed:
                if party_id is not None and party_id not in (trade.buyer, trade.seller):
                    continue
                if market_id is not None and trade.market_id != market_id:
                    continue
                if order_id is not None and order_id not in (
                    trade.buy_order,
                    trade.sell_order,
                ):
                    continue
                if exclude_trade_ids is not None and trade.id in exclude_trade_ids:
                    continue
                results.append(trade)
        return results



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: vega/governance.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import assets_pb2 as vega_dot_assets__pb2
from . import data_source_pb2 as vega_dot_data__source__pb2
from . import markets_pb2 as vega_dot_markets__pb2
from . import vega_pb2 as vega_dot_vega__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x15vega/governance.proto\x12\x04vega\x1a\x11vega/assets.proto\x1a\x16vega/data_source.proto\x1a\x12vega/markets.proto\x1a\x0fvega/vega.proto"a\n\x0bSpotProduct\x12\x1d\n\nbase_asset\x18\x01 \x01(\tR\tbaseAsset\x12\x1f\n\x0bquote_asset\x18\x02 \x01(\tR\nquoteAsset\x12\x12\n\x04name\x18\x03 \x01(\tR\x04name"\x95\x03\n\rFutureProduct\x12)\n\x10settlement_asset\x18\x01 \x01(\tR\x0fsettlementAsset\x12\x1d\n\nquote_name\x18\x02 \x01(\tR\tquoteName\x12i\n$data_source_spec_for_settlement_data\x18\x03 \x01(\x0b\x32\x1a.vega.DataSourceDefinitionR\x1f\x64\x61taSourceSpecForSettlementData\x12q\n(data_source_spec_for_trading_termination\x18\x04 \x01(\x0b\x32\x1a.vega.DataSourceDefinitionR#dataSourceSpecForTradingTermination\x12\\\n\x18\x64\x61ta_source_spec_binding\x18\x05 \x01(\x0b\x32#.vega.DataSourceSpecToFutureBindingR\x15\x64\x61taSourceSpecBinding"\xa4\x01\n\x17InstrumentConfiguration\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\x12\n\x04\x63ode\x18\x02 \x01(\tR\x04\x63ode\x12-\n\x06\x66uture\x18\x64 \x01(\x0b\x32\x13.vega.FutureProductH\x00R\x06\x66uture\x12\'\n\x04spot\x18\x65 \x01(\x0b\x32\x11.vega.SpotProductH\x00R\x04spotB\t\n\x07product"\x8d\x04\n\x1aNewSpotMarketConfiguration\x12=\n\ninstrument\x18\x01 \x01(\x0b\x32\x1d.vega.InstrumentConfigurationR\ninstrument\x12%\n\x0e\x64\x65\x63imal_places\x18\x02 \x01(\x04R\rdecimalPlaces\x12\x1a\n\x08metadata\x18\x03 \x03(\tR\x08metadata\x12_\n\x1bprice_monitoring_parameters\x18\x04 \x01(\x0b\x32\x1f.vega.PriceMonitoringParametersR\x19priceMonitoringParameters\x12S\n\x17target_stake_parameters\x18\x05 \x01(\x0b\x32\x1b.vega.TargetStakeParametersR\x15targetStakeParameters\x12\x31\n\x06simple\x18\x64 \x01(\x0b\x32\x17.vega.SimpleModelParamsH\x00R\x06simple\x12\x39\n\nlog_normal\x18\x65 \x01(\x0b\x32\x18.vega.LogNormalRiskModelH\x00R\tlogNormal\x12\x36\n\x17position_decimal_places\x18\x06 \x01(\x03R\x15positionDecimalPlacesB\x11\n\x0frisk_parameters"\x88\x06\n\x16NewMarketConfiguration\x12=\n\ninstrument\x18\x01 \x01(\x0b\x32\x1d.vega.InstrumentConfigurationR\ninstrument\x12%\n\x0e\x64\x65\x63imal_places\x18\x02 \x01(\x04R\rdecimalPlaces\x12\x1a\n\x08metadata\x18\x03 \x03(\tR\x08metadata\x12_\n\x1bprice_monitoring_parameters\x18\x04 \x01(\x0b\x32\x1f.vega.PriceMonitoringParametersR\x19priceMonitoringParameters\x12k\n\x1fliquidity_monitoring_parameters\x18\x05 \x01(\x0b\x32#.vega.LiquidityMonitoringParametersR\x1dliquidityMonitoringParameters\x12\x31\n\x06simple\x18\x64 \x01(\x0b\x32\x17.vega.SimpleModelParamsH\x00R\x06simple\x12\x39\n\nlog_normal\x18\x65 \x01(\x0b\x32\x18.vega.LogNormalRiskModelH\x00R\tlogNormal\x12\x36\n\x17position_decimal_places\x18\x06 \x01(\x03R\x15positionDecimalPlaces\x12$\n\x0elp_price_range\x18\x08 \x01(\tR\x0clpPriceRange\x12\x34\n\x16linear_slippage_factor\x18\t \x01(\tR\x14linearSlippageFactor\x12:\n\x19quadratic_slippage_factor\x18\n \x01(\tR\x17quadraticSlippageFactor\x12?\n\tsuccessor\x18\x0b \x01(\x0b\x32\x1c.vega.SuccessorConfigurationH\x01R\tsuccessor\x88\x01\x01\x42\x11\n\x0frisk_parametersB\x0c\n\n_successor"K\n\rNewSpotMarket\x12:\n\x07\x63hanges\x18\x01 \x01(\x0b\x32 .vega.NewSpotMarketConfigurationR\x07\x63hanges"z\n\x16SuccessorConfiguration\x12(\n\x10parent_market_id\x18\x01 \x01(\tR\x0eparentMarketId\x12\x36\n\x17insurance_pool_fraction\x18\x02 \x01(\tR\x15insurancePoolFraction"C\n\tNewMarket\x12\x36\n\x07\x63hanges\x18\x01 \x01(\x0b\x32\x1c.vega.NewMarketConfigurationR\x07\x63hanges"f\n\x0cUpdateMarket\x12\x1b\n\tmarket_id\x18\x01 \x01(\tR\x08marketId\x12\x39\n\x07\x63hanges\x18\x02 \x01(\x0b\x32\x1f.vega.UpdateMarketConfigurationR\x07\x63hanges"n\n\x10UpdateSpotMarket\x12\x1b\n\tmarket_id\x18\x01 \x01(\tR\x08marketId\x12=\n\x07\x63hanges\x18\x02 \x01(\x0b\x32#.vega.UpdateSpotMarketConfigurationR\x07\x63hanges"\xe3\x04\n\x19UpdateMarketConfiguration\x12\x43\n\ninstrument\x18\x01 \x01(\x0b\x32#.vega.UpdateInstrumentConfigurationR\ninstrument\x12\x1a\n\x08metadata\x18\x02 \x03(\tR\x08metadata\x12_\n\x1bprice_monitoring_parameters\x18\x03 \x01(\x0b\x32\x1f.vega.PriceMonitoringParametersR\x19priceMonitoringParameters\x12k\n\x1fliquidity_monitoring_parameters\x18\x04 \x01(\x0b\x32#.vega.LiquidityMonitoringParametersR\x1dliquidityMonitoringParameters\x12\x31\n\x06simple\x18\x64 \x01(\x0b\x32\x17.vega.SimpleModelParamsH\x00R\x06simple\x12\x39\n\nlog_normal\x18\x65 \x01(\x0b\x32\x18.vega.LogNormalRiskModelH\x00R\tlogNormal\x12$\n\x0elp_price_range\x18\x05 \x01(\tR\x0clpPriceRange\x12\x34\n\x16linear_slippage_factor\x18\x06 \x01(\tR\x14linearSlippageFactor\x12:\n\x19quadratic_slippage_factor\x18\x07 \x01(\tR\x17quadraticSlippageFactorB\x11\n\x0frisk_parameters"\xf2\x02\n\x1dUpdateSpotMarketConfiguration\x12\x1a\n\x08metadata\x18\x01 \x03(\tR\x08metadata\x12_\n\x1bprice_monitoring_parameters\x18\x02 \x01(\x0b\x32\x1f.vega.PriceMonitoringParametersR\x19priceMonitoringParameters\x12S\n\x17target_stake_parameters\x18\x03 \x01(\x0b\x32\x1b.vega.TargetStakeParametersR\x15targetStakeParameters\x12\x31\n\x06simple\x18\x64 \x01(\x0b\x32\x17.vega.SimpleModelParamsH\x00R\x06simple\x12\x39\n\nlog_normal\x18\x65 \x01(\x0b\x32\x18.vega.LogNormalRiskModelH\x00R\tlogNormalB\x11\n\x0frisk_parameters"s\n\x1dUpdateInstrumentConfiguration\x12\x12\n\x04\x63ode\x18\x01 \x01(\tR\x04\x63ode\x12\x33\n\x06\x66uture\x18\x64 \x01(\x0b\x32\x19.vega.UpdateFutureProductH\x00R\x06\x66utureB\t\n\x07product"\xf0\x02\n\x13UpdateFutureProduct\x12\x1d\n\nquote_name\x18\x01 \x01(\tR\tquoteName\x12i\n$data_source_spec_for_settlement_data\x18\x02 \x01(\x0b\x32\x1a.vega.DataSourceDefinitionR\x1f\x64\x61taSourceSpecForSettlementData\x12q\n(data_source_spec_for_trading_termination\x18\x03 \x01(\x0b\x32\x1a.vega.DataSourceDefinitionR#dataSourceSpecForTradingTermination\x12\\\n\x18\x64\x61ta_source_spec_binding\x18\x04 \x01(\x0b\x32#.vega.DataSourceSpecToFutureBindingR\x15\x64\x61taSourceSpecBinding"J\n\x16UpdateNetworkParameter\x12\x30\n\x07\x63hanges\x18\x01 \x01(\x0b\x32\x16.vega.NetworkParameterR\x07\x63hanges"8\n\x08NewAsset\x12,\n\x07\x63hanges\x18\x01 \x01(\x0b\x32\x12.vega.AssetDetailsR\x07\x63hanges"\\\n\x0bUpdateAsset\x12\x19\n\x08\x61sset_id\x18\x01 \x01(\tR\x07\x61ssetId\x12\x32\n\x07\x63hanges\x18\x02 \x01(\x0b\x32\x18.vega.AssetDetailsUpdateR\x07\x63hanges"\r\n\x0bNewFreeform"\x90\x06\n\rProposalTerms\x12+\n\x11\x63losing_timestamp\x18\x01 \x01(\x03R\x10\x63losingTimestamp\x12/\n\x13\x65nactment_timestamp\x18\x02 \x01(\x03R\x12\x65nactmentTimestamp\x12\x31\n\x14validation_timestamp\x18\x03 \x01(\x03R\x13validationTimestamp\x12\x39\n\rupdate_market\x18\x65 \x01(\x0b\x32\x12.vega.UpdateMarketH\x00R\x0cupdateMarket\x12\x30\n\nnew_market\x18\x66 \x01(\x0b\x32\x0f.vega.NewMarketH\x00R\tnewMarket\x12X\n\x18update_network_parameter\x18g \x01(\x0b\x32\x1c.vega.UpdateNetworkParameterH\x00R\x16updateNetworkParameter\x12-\n\tnew_asset\x18h \x01(\x0b\x32\x0e.vega.NewAssetH\x00R\x08newAsset\x12\x36\n\x0cnew_freeform\x18i \x01(\x0b\x32\x11.vega.NewFreeformH\x00R\x0bnewFreeform\x12\x36\n\x0cupdate_asset\x18j \x01(\x0b\x32\x11.vega.UpdateAssetH\x00R\x0bupdateAsset\x12=\n\x0fnew_spot_market\x18k \x01(\x0b\x32\x13.vega.NewSpotMarketH\x00R\rnewSpotMarket\x12\x46\n\x12update_spot_market\x18l \x01(\x0b\x32\x16.vega.UpdateSpotMarketH\x00R\x10updateSpotMarket\x12\x36\n\x0cnew_transfer\x18m \x01(\x0b\x32\x11.vega.NewTransferH\x00R\x0bnewTransfer\x12?\n\x0f\x63\x61ncel_transfer\x18n \x01(\x0b\x32\x14.vega.CancelTransferH\x00R\x0e\x63\x61ncelTransferB\x08\n\x06\x63hange"W\n\x11ProposalRationale\x12 \n\x0b\x64\x65scription\x18\x01 \x01(\tR\x0b\x64\x65scription\x12\x14\n\x05title\x18\x04 \x01(\tR\x05titleJ\x04\x08\x02\x10\x03J\x04\x08\x03\x10\x04"\x86\x03\n\x0eGovernanceData\x12*\n\x08proposal\x18\x01 \x01(\x0b\x32\x0e.vega.ProposalR\x08proposal\x12\x1c\n\x03yes\x18\x02 \x03(\x0b\x32\n.vega.VoteR\x03yes\x12\x1a\n\x02no\x18\x03 \x03(\x0b\x32\n.vega.VoteR\x02no\x12?\n\tyes_party\x18\x04 \x03(\x0b\x32".vega.GovernanceData.YesPartyEntryR\x08yesParty\x12<\n\x08no_party\x18\x05 \x03(\x0b\x32!.vega.GovernanceData.NoPartyEntryR\x07noParty\x1aG\n\rYesPartyEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12 \n\x05value\x18\x02 \x01(\x0b\x32\n.vega.VoteR\x05value:\x02\x38\x01\x1a\x46\n\x0cNoPartyEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12 \n\x05value\x18\x02 \x01(\x0b\x32\n.vega.VoteR\x05value:\x02\x38\x01"\x9a\x07\n\x08Proposal\x12\x0e\n\x02id\x18\x01 \x01(\tR\x02id\x12\x1c\n\treference\x18\x02 \x01(\tR\treference\x12\x19\n\x08party_id\x18\x03 \x01(\tR\x07partyId\x12*\n\x05state\x18\x04 \x01(\x0e\x32\x14.vega.Proposal.StateR\x05state\x12\x1c\n\ttimestamp\x18\x05 \x01(\x03R\ttimestamp\x12)\n\x05terms\x18\x06 \x01(\x0b\x32\x13.vega.ProposalTermsR\x05terms\x12\x30\n\x06reason\x18\x07 \x01(\x0e\x32\x13.vega.ProposalErrorH\x00R\x06reason\x88\x01\x01\x12(\n\rerror_details\x18\x08 \x01(\tH\x01R\x0c\x65rrorDetails\x88\x01\x01\x12\x35\n\trationale\x18\t \x01(\x0b\x32\x17.vega.ProposalRationaleR\trationale\x12\x35\n\x16required_participation\x18\n \x01(\tR\x15requiredParticipation\x12+\n\x11required_majority\x18\x0b \x01(\tR\x10requiredMajority\x12^\n)required_liquidity_provider_participation\x18\x0c \x01(\tH\x02R&requiredLiquidityProviderParticipation\x88\x01\x01\x12T\n$required_liquidity_provider_majority\x18\r \x01(\tH\x03R!requiredLiquidityProviderMajority\x88\x01\x01"\xae\x01\n\x05State\x12\x15\n\x11STATE_UNSPECIFIED\x10\x00\x12\x10\n\x0cSTATE_FAILED\x10\x01\x12\x0e\n\nSTATE_OPEN\x10\x02\x12\x10\n\x0cSTATE_PASSED\x10\x03\x12\x12\n\x0eSTATE_REJECTED\x10\x04\x12\x12\n\x0eSTATE_DECLINED\x10\x05\x12\x11\n\rSTATE_ENACTED\x10\x06\x12\x1f\n\x1bSTATE_WAITING_FOR_NODE_VOTE\x10\x07\x42\t\n\x07_reasonB\x10\n\x0e_error_detailsB,\n*_required_liquidity_provider_participationB\'\n%_required_liquidity_provider_majority"\x91\x03\n\x04Vote\x12\x19\n\x08party_id\x18\x01 \x01(\tR\x07partyId\x12&\n\x05value\x18\x02 \x01(\x0e\x32\x10.vega.Vote.ValueR\x05value\x12\x1f\n\x0bproposal_id\x18\x03 \x01(\tR\nproposalId\x12\x1c\n\ttimestamp\x18\x04 \x01(\x03R\ttimestamp\x12\x43\n\x1etotal_governance_token_balance\x18\x05 \x01(\tR\x1btotalGovernanceTokenBalance\x12\x41\n\x1dtotal_governance_token_weight\x18\x06 \x01(\tR\x1atotalGovernanceTokenWeight\x12\x42\n\x1etotal_equity_like_share_weight\x18\x07 \x01(\tR\x1atotalEquityLikeShareWeight";\n\x05Value\x12\x15\n\x11VALUE_UNSPECIFIED\x10\x00\x12\x0c\n\x08VALUE_NO\x10\x01\x12\r\n\tVALUE_YES\x10\x02"M\n\x0e\x43\x61ncelTransfer\x12;\n\x07\x63hanges\x18\x01 \x01(\x0b\x32!.vega.CancelTransferConfigurationR\x07\x63hanges">\n\x1b\x43\x61ncelTransferConfiguration\x12\x1f\n\x0btransfer_id\x18\x01 \x01(\tR\ntransferId"G\n\x0bNewTransfer\x12\x38\n\x07\x63hanges\x18\x01 \x01(\x0b\x32\x1e.vega.NewTransferConfigurationR\x07\x63hanges"\xd9\x03\n\x18NewTransferConfiguration\x12\x32\n\x0bsource_type\x18\x01 \x01(\x0e\x32\x11.vega.AccountTypeR\nsourceType\x12\x16\n\x06source\x18\x02 \x01(\tR\x06source\x12\x41\n\rtransfer_type\x18\x03 \x01(\x0e\x32\x1c.vega.GovernanceTransferTypeR\x0ctransferType\x12\x16\n\x06\x61mount\x18\x04 \x01(\tR\x06\x61mount\x12\x14\n\x05\x61sset\x18\x05 \x01(\tR\x05\x61sset\x12.\n\x13\x66raction_of_balance\x18\x06 \x01(\tR\x11\x66ractionOfBalance\x12<\n\x10\x64\x65stination_type\x18\x07 \x01(\x0e\x32\x11.vega.AccountTypeR\x0f\x64\x65stinationType\x12 \n\x0b\x64\x65stination\x18\x08 \x01(\tR\x0b\x64\x65stination\x12/\n\x07one_off\x18\x65 \x01(\x0b\x32\x14.vega.OneOffTransferH\x00R\x06oneOff\x12\x37\n\trecurring\x18\x66 \x01(\x0b\x32\x17.vega.RecurringTransferH\x00R\trecurringB\x06\n\x04kind"/\n\x0eOneOffTransfer\x12\x1d\n\ndeliver_on\x18\x01 \x01(\x03R\tdeliverOn"d\n\x11RecurringTransfer\x12\x1f\n\x0bstart_epoch\x18\x01 \x01(\x04R\nstartEpoch\x12 \n\tend_epoch\x18\x02 \x01(\x04H\x00R\x08\x65ndEpoch\x88\x01\x01\x42\x0c\n\n_end_epoch*\xc9\x10\n\rProposalError\x12\x1e\n\x1aPROPOSAL_ERROR_UNSPECIFIED\x10\x00\x12&\n"PROPOSAL_ERROR_CLOSE_TIME_TOO_SOON\x10\x01\x12&\n"PROPOSAL_ERROR_CLOSE_TIME_TOO_LATE\x10\x02\x12&\n"PROPOSAL_ERROR_ENACT_TIME_TOO_SOON\x10\x03\x12&\n"PROPOSAL_ERROR_ENACT_TIME_TOO_LATE\x10\x04\x12&\n"PROPOSAL_ERROR_INSUFFICIENT_TOKENS\x10\x05\x12.\n*PROPOSAL_ERROR_INVALID_INSTRUMENT_SECURITY\x10\x06\x12\x1d\n\x19PROPOSAL_ERROR_NO_PRODUCT\x10\x07\x12&\n"PROPOSAL_ERROR_UNSUPPORTED_PRODUCT\x10\x08\x12"\n\x1ePROPOSAL_ERROR_NO_TRADING_MODE\x10\x0b\x12+\n\'PROPOSAL_ERROR_UNSUPPORTED_TRADING_MODE\x10\x0c\x12)\n%PROPOSAL_ERROR_NODE_VALIDATION_FAILED\x10\r\x12.\n*PROPOSAL_ERROR_MISSING_BUILTIN_ASSET_FIELD\x10\x0e\x12\x31\n-PROPOSAL_ERROR_MISSING_ERC20_CONTRACT_ADDRESS\x10\x0f\x12 \n\x1cPROPOSAL_ERROR_INVALID_ASSET\x10\x10\x12*\n&PROPOSAL_ERROR_INCOMPATIBLE_TIMESTAMPS\x10\x11\x12%\n!PROPOSAL_ERROR_NO_RISK_PARAMETERS\x10\x12\x12\x30\n,PROPOSAL_ERROR_NETWORK_PARAMETER_INVALID_KEY\x10\x13\x12\x32\n.PROPOSAL_ERROR_NETWORK_PARAMETER_INVALID_VALUE\x10\x14\x12\x36\n2PROPOSAL_ERROR_NETWORK_PARAMETER_VALIDATION_FAILED\x10\x15\x12\x35\n1PROPOSAL_ERROR_OPENING_AUCTION_DURATION_TOO_SMALL\x10\x16\x12\x35\n1PROPOSAL_ERROR_OPENING_AUCTION_DURATION_TOO_LARGE\x10\x17\x12/\n+PROPOSAL_ERROR_COULD_NOT_INSTANTIATE_MARKET\x10\x19\x12)\n%PROPOSAL_ERROR_INVALID_FUTURE_PRODUCT\x10\x1a\x12)\n%PROPOSAL_ERROR_INVALID_RISK_PARAMETER\x10\x1e\x12\x31\n-PROPOSAL_ERROR_MAJORITY_THRESHOLD_NOT_REACHED\x10\x1f\x12\x36\n2PROPOSAL_ERROR_PARTICIPATION_THRESHOLD_NOT_REACHED\x10 \x12(\n$PROPOSAL_ERROR_INVALID_ASSET_DETAILS\x10!\x12\x1f\n\x1bPROPOSAL_ERROR_UNKNOWN_TYPE\x10"\x12.\n*PROPOSAL_ERROR_UNKNOWN_RISK_PARAMETER_TYPE\x10#\x12#\n\x1fPROPOSAL_ERROR_INVALID_FREEFORM\x10$\x12\x31\n-PROPOSAL_ERROR_INSUFFICIENT_EQUITY_LIKE_SHARE\x10%\x12!\n\x1dPROPOSAL_ERROR_INVALID_MARKET\x10&\x12\x31\n-PROPOSAL_ERROR_TOO_MANY_MARKET_DECIMAL_PLACES\x10\'\x12\x35\n1PROPOSAL_ERROR_TOO_MANY_PRICE_MONITORING_TRIGGERS\x10(\x12/\n+PROPOSAL_ERROR_ERC20_ADDRESS_ALREADY_IN_USE\x10)\x12-\n)PROPOSAL_ERROR_LP_PRICE_RANGE_NONPOSITIVE\x10*\x12+\n\'PROPOSAL_ERROR_LP_PRICE_RANGE_TOO_LARGE\x10+\x12\x36\n2PROPOSAL_ERROR_LINEAR_SLIPPAGE_FACTOR_OUT_OF_RANGE\x10,\x12\x39\n5PROPOSAL_ERROR_QUADRATIC_SLIPPAGE_FACTOR_OUT_OF_RANGE\x10-\x12\x1f\n\x1bPROPOSAL_ERROR_INVALID_SPOT\x10.\x12(\n$PROPOSAL_ERROR_SPOT_PRODUCT_DISABLED\x10/\x12+\n\'PROPOSAL_ERROR_INVALID_SUCCESSOR_MARKET\x10\x30\x12\x36\n2PROPOSAL_ERROR_GOVERNANCE_TRANSFER_PROPOSAL_FAILED\x10\x31\x12\x37\n3PROPOSAL_ERROR_GOVERNANCE_TRANSFER_PROPOSAL_INVALID\x10\x32\x12>\n:PROPOSAL_ERROR_GOVERNANCE_CANCEL_TRANSFER_PROPOSAL_INVALID\x10\x33*\x99\x01\n\x16GovernanceTransferType\x12(\n$GOVERNANCE_TRANSFER_TYPE_UNSPECIFIED\x10\x00\x12+\n\'GOVERNANCE_TRANSFER_TYPE_ALL_OR_NOTHING\x10\x01\x12(\n$GOVERNANCE_TRANSFER_TYPE_BEST_EFFORT\x10\x02\x42\'Z%code.vegaprotocol.io/vega/protos/vegab\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "vega.governance_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    DESCRIPTOR._serialized_options = b"Z%code.vegaprotocol.io/vega/protos/vega"
    _GOVERNANCEDATA_YESPARTYENTRY._options = None
    _GOVERNANCEDATA_YESPARTYENTRY._serialized_options = b"8\001"
    _GOVERNANCEDATA_NOPARTYENTRY._options = None
    _GOVERNANCEDATA_NOPARTYENTRY._serialized_options = b"8\001"
    _globals["_PROPOSALERROR"]._serialized_start = 7738
    _globals["_PROPOSALERROR"]._serialized_end = 9859
    _globals["_GOVERNANCETRANSFERTYPE"]._serialized_start = 9862
    _globals["_GOVERNANCETRANSFERTYPE"]._serialized_end = 10015
    _globals["_SPOTPRODUCT"]._serialized_start = 111
    _globals["_SPOTPRODUCT"]._serialized_end = 208
    _globals["_FUTUREPRODUCT"]._serialized_start = 211
    _globals["_FUTUREPRODUCT"]._serialized_end = 616
    _globals["_INSTRUMENTCONFIGURATION"]._serialized_start = 619
    _globals["_INSTRUMENTCONFIGURATION"]._serialized_end = 783
    _globals["_NEWSPOTMARKETCONFIGURATION"]._serialized_start = 786
    _globals["_NEWSPOTMARKETCONFIGURATION"]._serialized_end = 1311
    _globals["_NEWMARKETCONFIGURATION"]._serialized_start = 1314
    _globals["_NEWMARKETCONFIGURATION"]._serialized_end = 2090
    _globals["_NEWSPOTMARKET"]._serialized_start = 2092
    _globals["_NEWSPOTMARKET"]._serialized_end = 2167
    _globals["_SUCCESSORCONFIGURATION"]._serialized_start = 2169
    _globals["_SUCCESSORCONFIGURATION"]._serialized_end = 2291
    _globals["_NEWMARKET"]._serialized_start = 2293
    _globals["_NEWMARKET"]._serialized_end = 2360
    _globals["_UPDATEMARKET"]._serialized_start = 2362
    _globals["_UPDATEMARKET"]._serialized_end = 2464
    _globals["_UPDATESPOTMARKET"]._serialized_start = 2466
    _globals["_UPDATESPOTMARKET"]._serialized_end = 2576
    _globals["_UPDATEMARKETCONFIGURATION"]._serialized_start = 2579
    _globals["_UPDATEMARKETCONFIGURATION"]._serialized_end = 3190
    _globals["_UPDATESPOTMARKETCONFIGURATION"]._serialized_start = 3193
    _globals["_UPDATESPOTMARKETCONFIGURATION"]._serialized_end = 3563
    _globals["_UPDATEINSTRUMENTCONFIGURATION"]._serialized_start = 3565
    _globals["_UPDATEINSTRUMENTCONFIGURATION"]._serialized_end = 3680
    _globals["_UPDATEFUTUREPRODUCT"]._serialized_start = 3683
    _globals["_UPDATEFUTUREPRODUCT"]._serialized_end = 4051
    _globals["_UPDATENETWORKPARAMETER"]._serialized_start = 4053
    _globals["_UPDATENETWORKPARAMETER"]._serialized_end = 4127
    _globals["_NEWASSET"]._serialized_start = 4129
    _globals["_NEWASSET"]._serialized_end = 4185
    _globals["_UPDATEASSET"]._serialized_start = 4187
    _globals["_UPDATEASSET"]._serialized_end = 4279
    _globals["_NEWFREEFORM"]._serialized_start = 4281
    _globals["_NEWFREEFORM"]._serialized_end = 4294
    _globals["_PROPOSALTERMS"]._serialized_start = 4297
    _globals["_PROPOSALTERMS"]._serialized_end = 5081
    _globals["_PROPOSALRATIONALE"]._serialized_start = 5083
    _globals["_PROPOSALRATIONALE"]._serialized_end = 5170
    _globals["_GOVERNANCEDATA"]._serialized_start = 5173
    _globals["_GOVERNANCEDATA"]._serialized_end = 5563
    _globals["_GOVERNANCEDATA_YESPARTYENTRY"]._serialized_start = 5420
    _globals["_GOVERNANCEDATA_YESPARTYENTRY"]._serialized_end = 5491
    _globals["_GOVERNANCEDATA_NOPARTYENTRY"]._serialized_start = 5493
    _globals["_GOVERNANCEDATA_NOPARTYENTRY"]._serialized_end = 5563
    _globals["_PROPOSAL"]._serialized_start = 5566
    _globals["_PROPOSAL"]._serialized_end = 6488
    _globals["_PROPOSAL_STATE"]._serialized_start = 6198
    _globals["_PROPOSAL_STATE"]._serialized_end = 6372
    _globals["_VOTE"]._serialized_start = 6491
    _globals["_VOTE"]._serialized_end = 6892
    _globals["_VOTE_VALUE"]._serialized_start = 6833
    _globals["_VOTE_VALUE"]._serialized_end = 6892
    _globals["_CANCELTRANSFER"]._serialized_start = 6894
    _globals["_CANCELTRANSFER"]._serialized_end = 6971
    _globals["_CANCELTRANSFERCONFIGURATION"]._serialized_start = 6973
    _globals["_CANCELTRANSFERCONFIGURATION"]._serialized_end = 7035
    _globals["_NEWTRANSFER"]._serialized_start = 7037
    _globals["_NEWTRANSFER"]._serialized_end = 7108
    _globals["_NEWTRANSFERCONFIGURATION"]._serialized_start = 7111
    _globals["_NEWTRANSFERCONFIGURATION"]._serialized_end = 7584
    _globals["_ONEOFFTRANSFER"]._serialized_start = 7586
    _globals["_ONEOFFTRANSFER"]._serialized_end = 7633
    _globals["_RECURRINGTRANSFER"]._serialized_start = 7635
    _globals["_RECURRINGTRANSFER"]._serialized_end = 7735
# @@protoc_insertion_point(module_scope)



from concurrent.futures import ThreadPoolExecutor
from enum import Enum, auto
from io import BufferedWriter
from typing import Any
from logging import getLogger

import os
import dotenv
import numpy as np
from nacl.encoding import HexEncoder
from nacl.signing import SigningKey
from typing import Optional
from vega_sim.grpc.client import VegaCoreClient
from vega_sim.wallet.base import Wallet

import vega_sim.proto.vega.api.v1.core_pb2 as core_proto
import vega_sim.proto.vega.commands.v1.transaction_pb2 as transaction_proto
import vega_sim.proto.vega.commands.v1.signature_pb2 as signature_proto
from vega_sim.wallet.vega_wallet import VegaWallet
from vega_sim.wallet.base import DEFAULT_WALLET_NAME

logger = getLogger(__name__)


class SlimWallet(Wallet):
    def __init__(
        self,
        core_client: VegaCoreClient,
        full_wallet: Optional[VegaWallet] = None,
        store_transactions: bool = False,
        log_dir: Optional[str] = None,
    ):
        """Creates a wallet to running key generation internally
        and directly sending transactions to the Core node

        Args:
            core_client:
                VegaCoreClient, client for vega core interaction
            full_wallet:
                Optional[VegaWallet], optional full wallet backing. This will be used
                    for creating keypairs instead of generating them locally, making it
                    possible to connect to the Console
            store_transactions:
                bool, default False, If True will store every transaction sent into
                    a file, allowing replay of the chain without going through full
                    logic of actors etc
        """
        self.core_client = core_client
        self.keys = {}
        self.pub_keys = {}
        self.idx = 0
        self.nonce_idx = 0
        self.pool = ThreadPoolExecutor(max_workers=5)

        self.height_update_frequency = 500
        self.remaining_until_height_update = 0
        self.block_height = None

        self.vega_wallet = full_wallet

        self.store_transactions = store_transactions
        self.log_dir = log_dir

        # If it turns out that customising these is useful it's trivial to
        # make a parameter
        self.num_sigs_to_create = 1000000
        self.num_nonces_to_create = 1000000

        self._create_nonces()
        self._create_sigs()

        dotenv.load_dotenv()
        self.vega_default_wallet_name = os.environ.get(
            "VEGA_DEFAULT_WALLET_NAME", DEFAULT_WALLET_NAME
        )

    def _create_sigs(self):
        self.sigs = [os.urandom(6).hex() for _ in range(self.num_sigs_to_create)]

    def _create_nonces(self):
        self.nonces = np.random.randint(0, 100000, size=self.num_nonces_to_create)

    def _next_sig(self) -> bytes:
        self.idx += 1
        if self.idx >= self.num_sigs_to_create:
            self.idx = 0
            self._create_sigs()
        return self.sigs[self.idx]

    def _next_nonce(self) -> int:
        self.nonce_idx += 1
        if self.nonce_idx >= self.num_nonces_to_create:
            self.nonce_idx = 0
            self._create_nonces()
        return self.nonces[self.nonce_idx]

    def create_key(
        self,
        name: str,
        wallet_name: Optional[str] = None,
    ) -> str:
        """Generates a new wallet key in the given vega service.

        Args:
            name:
                str, The name to use for the wallet key
            wallet_name:
                str, The wallet to use if not the default
        Returns:
            str, login token to use in authenticated requests
        """
        wallet_name = (
            wallet_name if wallet_name is not None else self.vega_default_wallet_name
        )

        if wallet_name not in self.keys:
            self.keys[wallet_name] = {}
        if wallet_name not in self.pub_keys:
            self.pub_keys[wallet_name] = {}

        if self.vega_wallet is None:
            self.keys[wallet_name][name] = SigningKey.generate()
            self.pub_keys[wallet_name][name] = (
                self.keys[wallet_name][name]
                .verify_key.encode(encoder=HexEncoder)
                .decode()
            )

        else:
            self.vega_wallet.create_key(
                name=name,
                wallet_name=wallet_name,
            )
            self.pub_keys[wallet_name][name] = self.vega_wallet.public_key(
                name, wallet_name
            )

    def submit_transaction(
        self,
        key_name: str,
        transaction: Any,
        transaction_type: str,
        wallet_name: Optional[str] = None,
    ):
        # if self.remaining_until_height_update <= 0:
        #     self.block_height = self.core_client.LastBlockHeight(
        #         core_proto.LastBlockHeightRequest()
        #     ).height
        #     self.remaining_until_height_update = self.height_update_frequency

        pub_key = self.public_key(name=key_name, wallet_name=wallet_name)

        transaction_info = {transaction_type: transaction}
        input_data = transaction_proto.InputData(
            nonce=self._next_nonce(),  # block_height=self.block_height,
            **transaction_info
        )

        serialised = input_data.SerializeToString()

        trans = transaction_proto.Transaction(
            input_data=serialised,
            signature=signature_proto.Signature(
                value=self._next_sig(),
                algo="vega/ed25519",
                version=1,
            ),
            pub_key=pub_key,
            version=3,
            pow=transaction_proto.ProofOfWork(
                tid=None,
                nonce=0,
            ),
        )
        request = core_proto.SubmitTransactionRequest(
            tx=trans, type=core_proto.SubmitTransactionRequest.Type.TYPE_ASYNC
        )

        submit_future = self.core_client.SubmitTransaction.future(request)
        self.pool.submit(lambda: submit_future.result())
        self.remaining_until_height_update -= 1

    def submit_raw_transaction(self, transaction: core_proto.SubmitTransactionRequest):
        submit_future = self.core_client.SubmitTransaction.future(transaction)
        self.pool.submit(lambda: submit_future.result())

    def public_key(self, name: str, wallet_name: Optional[str] = None) -> str:
        """Return a public key for the given wallet name and key name.

        Args:
            name (str):
                Name of the key.
            wallet_name (str):
                Name of the wallet. Defaults to None.

        Returns:
            str, public key
        """

        if wallet_name is None:
            return self.pub_keys[self.vega_default_wallet_name][name]
        else:
            return self.pub_keys[wallet_name][name]

    def stop(self):
        self.pool.shutdown(wait=True)



import copy
import os
from typing import Optional, Dict, List

import pandas as pd
from numpy.random import RandomState

import vega_sim.proto.vega as vega_protos
from vega_sim.api.market import MarketConfig, Successor
from vega_sim.environment.agent import StateAgentWithWallet
from vega_sim.null_service import VegaServiceNull
from vega_sim.proto.vega import markets as markets_protos
from vega_sim.service import VegaService


class FuzzingAgent(StateAgentWithWallet):
    NAME_BASE = "fuzzing_agent"

    # Set the memory which all instances can modify
    MEMORY = {
        key: []
        for key in (
            "TRADING_MODE",
            "COMMAND",
            "TYPE",
            "SIDE",
            "TIME_IN_FORCE",
            "PEGGED_REFERENCE",
            "POST_ONLY",
            "REDUCE_ONLY",
        )
    }
    # Set the output flag which all instances can modify
    OUTPUTTED = False

    def __init__(
        self,
        key_name: str,
        market_name: str,
        asset_name: str,
        tag: Optional[str] = None,
        wallet_name: Optional[str] = None,
        state_update_freq: Optional[int] = None,
        initial_asset_mint: float = 1e9,
        random_state: Optional[RandomState] = None,
        output_plot_on_finalise: bool = False,
    ):
        super().__init__(
            key_name=key_name,
            tag=tag,
            wallet_name=wallet_name,
            state_update_freq=state_update_freq,
        )

        self.market_name = market_name
        self.asset_name = asset_name
        self.initial_asset_mint = initial_asset_mint
        self.random_state = random_state if random_state is not None else RandomState()
        self.output_plot_on_finalise = output_plot_on_finalise

    def initialise(
        self, vega: VegaServiceNull, create_key: bool = True, mint_key: bool = True
    ):
        super().initialise(vega, create_key)

        self.market_id = self.vega.find_market_id(name=self.market_name)

        # Get asset id
        self.asset_id = self.vega.find_asset_id(symbol=self.asset_name)
        if mint_key:
            # Top up asset
            self.vega.mint(
                key_name=self.key_name,
                asset=self.asset_id,
                amount=self.initial_asset_mint,
                wallet_name=self.wallet_name,
            )

        self.vega.wait_fn(5)

    def step(self, vega_state):
        self.live_orders = self.vega.orders_for_party_from_feed(
            key_name=self.key_name,
            wallet_name=self.wallet_name,
            market_id=self.market_id,
            live_only=True,
        )

        self.curr_price = vega_state.market_state[self.market_id].midprice

        submissions = [self.create_fuzzed_submission(vega_state) for _ in range(20)]
        amendments = [self.create_fuzzed_amendment(vega_state) for _ in range(10)]
        cancellations = [self.create_fuzzed_cancellation(vega_state) for _ in range(1)]

        self.vega.submit_instructions(
            key_name=self.key_name,
            wallet_name=self.wallet_name,
            submissions=submissions,
            amendments=amendments,
            cancellations=cancellations,
        )

    def create_fuzzed_cancellation(self, vega_state):
        order_id = self._select_order_id()

        return self.vega.create_order_cancellation(
            order_id=order_id,
            market_id=self.market_id,
        )

    def create_fuzzed_amendment(self, vega_state):
        order_id = self._select_order_id()

        return self.vega.create_order_amendment(
            order_id=order_id,
            market_id=self.market_id,
            size_delta=self.random_state.normal(loc=0, scale=5),
            time_in_force=self.random_state.choice(
                a=[
                    "TIME_IN_FORCE_UNSPECIFIED",
                    "TIME_IN_FORCE_GTC",
                    "TIME_IN_FORCE_GTT",
                    "TIME_IN_FORCE_GFN",
                    "TIME_IN_FORCE_GFA",
                    "TIME_IN_FORCE_FOK",
                    "TIME_IN_FORCE_IOC",
                ]
            ),
            price=self.random_state.choice(
                a=[None, self.random_state.normal(loc=self.curr_price, scale=20)]
            ),
            expires_at=int(
                self.vega.get_blockchain_time()
                + self.random_state.normal(loc=60, scale=60) * 1e9
            ),
            pegged_reference=self.random_state.choice(
                a=[
                    "PEGGED_REFERENCE_UNSPECIFIED",
                    "PEGGED_REFERENCE_MID",
                    "PEGGED_REFERENCE_BEST_BID",
                    "PEGGED_REFERENCE_BEST_ASK",
                ],
                p=[0.5, 0.5 / 3, 0.5 / 3, 0.5 / 3],
            ),
            pegged_offset=self.random_state.normal(loc=0, scale=10),
        )

    def create_fuzzed_submission(self, vega_state):
        FuzzingAgent.MEMORY["TRADING_MODE"].append(
            markets_protos.Market.TradingMode.Name(
                vega_state.market_state[self.market_id].trading_mode
            )
        )
        FuzzingAgent.MEMORY["COMMAND"].append("ORDER_SUBMISSION")
        FuzzingAgent.MEMORY["SIDE"].append(
            self.random_state.choice(
                a=[
                    "SIDE_UNSPECIFIED",
                    "SIDE_BUY",
                    "SIDE_SELL",
                ]
            )
        )
        FuzzingAgent.MEMORY["TYPE"].append(
            self.random_state.choice(
                a=[
                    "TYPE_UNSPECIFIED",
                    "TYPE_MARKET",
                    "TYPE_LIMIT",
                ]
            )
        )
        FuzzingAgent.MEMORY["TIME_IN_FORCE"].append(
            self.random_state.choice(
                a=[
                    "TIME_IN_FORCE_UNSPECIFIED",
                    "TIME_IN_FORCE_GTC",
                    "TIME_IN_FORCE_GTT",
                    "TIME_IN_FORCE_GFN",
                    "TIME_IN_FORCE_GFA",
                    "TIME_IN_FORCE_FOK",
                    "TIME_IN_FORCE_IOC",
                ]
            )
        )
        FuzzingAgent.MEMORY["PEGGED_REFERENCE"].append(
            self.random_state.choice(
                a=[
                    "PEGGED_REFERENCE_UNSPECIFIED",
                    "PEGGED_REFERENCE_MID",
                    "PEGGED_REFERENCE_BEST_BID",
                    "PEGGED_REFERENCE_BEST_ASK",
                ],
                p=[0.5, 0.5 / 3, 0.5 / 3, 0.5 / 3],
            )
        )
        FuzzingAgent.MEMORY["REDUCE_ONLY"].append(
            self.random_state.choice(
                a=[
                    True,
                    False,
                ],
            )
        )
        FuzzingAgent.MEMORY["POST_ONLY"].append(
            self.random_state.choice(
                a=[
                    True,
                    False,
                ],
            )
        )

        return self.vega.create_order_submission(
            market_id=self.market_id,
            side=FuzzingAgent.MEMORY["SIDE"][-1],
            size=self.random_state.poisson(lam=10),
            order_type=FuzzingAgent.MEMORY["TYPE"][-1],
            time_in_force=FuzzingAgent.MEMORY["TIME_IN_FORCE"][-1],
            price=self.random_state.choice(
                a=[None, self.random_state.normal(loc=self.curr_price, scale=10)]
            ),
            expires_at=int(
                self.vega.get_blockchain_time()
                + self.random_state.normal(loc=60, scale=60) * 1e9
            ),
            pegged_reference=FuzzingAgent.MEMORY["PEGGED_REFERENCE"][-1],
            pegged_offset=self.random_state.normal(loc=0, scale=10),
            reduce_only=FuzzingAgent.MEMORY["REDUCE_ONLY"][-1],
            post_only=FuzzingAgent.MEMORY["POST_ONLY"][-1],
        )

    def _select_order_id(self):
        if self.live_orders != {}:
            order_key = self.random_state.choice(list(self.live_orders.keys()))
            order = self.live_orders.get(order_key)
            return order.id if order is not None else None
        else:
            return None

    def finalise(self):
        if self.output_plot_on_finalise:
            if not FuzzingAgent.OUTPUTTED:
                import plotly.express as px

                FuzzingAgent.OUTPUTTED = True
                df = pd.DataFrame.from_dict(FuzzingAgent.MEMORY)
                df = (
                    df.groupby(list(FuzzingAgent.MEMORY.keys()))
                    .size()
                    .reset_index()
                    .rename(columns={0: "count"})
                )

                range_color = (10, 5000)
                custom_color_scale = [
                    [0, "red"],
                    [range_color[0] / range_color[1], "red"],
                    [range_color[0] / range_color[1], "yellow"],
                    [1, "green"],
                ]

                fig = px.treemap(
                    df,
                    title="Fuzzed Trader Coverage",
                    path=list(self.__class__.MEMORY.keys()),
                    values="count",
                    color="count",
                    color_continuous_scale=custom_color_scale,
                    range_color=range_color,
                )
                fig.update_traces(marker=dict(cornerradius=5))

                if not os.path.exists("fuzz_plots"):
                    os.mkdir("fuzz_plots")

                fig.write_html("fuzz_plots/coverage.html")


class RiskyMarketOrderTrader(StateAgentWithWallet):
    NAME_BASE = "risky_market_order_trader"

    def __init__(
        self,
        key_name: str,
        market_name: str,
        asset_name: str,
        tag: Optional[str] = None,
        wallet_name: Optional[str] = None,
        state_update_freq: Optional[int] = None,
        random_state: Optional[RandomState] = None,
        side: str = "SIDE_BUY",
        size_factor: float = 0.6,
        initial_asset_mint: float = 1e5,
        step_bias: float = 0.05,
    ):
        super().__init__(key_name, tag, wallet_name, state_update_freq)

        self.market_name = market_name
        self.asset_name = asset_name

        self.size_factor = size_factor
        self.side = side
        self.random_state = random_state if random_state is not None else RandomState()
        self.initial_asset_mint = initial_asset_mint
        self.step_bias = step_bias

        self.close_outs = 0

    def initialise(self, vega: VegaServiceNull, create_key: bool = True, mint_key=True):
        super().initialise(vega, create_key)

        self.market_id = self.vega.find_market_id(name=self.market_name)

        # Get asset id
        self.asset_id = self.vega.find_asset_id(symbol=self.asset_name)
        if mint_key:
            # Top up asset
            self.vega.mint(
                key_name=self.key_name,
                asset=self.asset_id,
                amount=self.initial_asset_mint,
                wallet_name=self.wallet_name,
            )

        self.vega.wait_fn(5)

    def step(self, vega_state):
        account = self.vega.party_account(
            key_name=self.key_name,
            wallet_name=self.wallet_name,
            market_id=self.market_id,
            asset_id=self.asset_id,
        )

        if account.general + account.margin == 0:
            self.vega.mint(
                key_name=self.key_name,
                wallet_name=self.wallet_name,
                amount=self.initial_asset_mint,
                asset=self.asset_id,
            )
            self.close_outs += 1
            self.commitment_amount = 0
            return

        if (
            vega_state.market_state[self.market_id].trading_mode
            != markets_protos.Market.TradingMode.TRADING_MODE_CONTINUOUS
        ):
            return
        if self.random_state.rand() > self.step_bias:
            return

        midprice = vega_state.market_state[self.market_id].midprice

        if account.general > 0:
            add_to_margin = max(
                (account.general + account.margin) * self.size_factor - account.margin,
                0,
            )

            risk_factors = self.vega.get_risk_factors(market_id=self.market_id)
            risk_factor = (
                risk_factors.long if self.side == "SIDE_BUY" else risk_factors.short
            )

            size = add_to_margin / (midprice * risk_factor + 1e-20)

            try:
                self.vega.submit_market_order(
                    trading_key=self.key_name,
                    trading_wallet=self.wallet_name,
                    market_id=self.market_id,
                    side=self.side,
                    volume=size,
                    wait=False,
                )
            except Exception as e:
                import pdb

                pdb.set_trace()
                print(f"There was an error {e}")


class RiskySimpleLiquidityProvider(StateAgentWithWallet):
    NAME_BASE = "risky_simple_liquidity_provider"

    def __init__(
        self,
        key_name: str,
        market_name: str,
        asset_name: str,
        tag: Optional[str] = None,
        wallet_name: Optional[str] = None,
        state_update_freq: Optional[int] = None,
        random_state: Optional[RandomState] = None,
        commitment_factor: float = 0.5,
        initial_asset_mint: float = 1e5,
        step_bias: float = 0.05,
    ):
        super().__init__(key_name, tag, wallet_name, state_update_freq)

        self.market_name = market_name
        self.asset_name = asset_name

        self.commitment_factor = commitment_factor
        self.random_state = random_state if random_state is not None else RandomState()
        self.initial_asset_mint = initial_asset_mint
        self.step_bias = step_bias

        self.commitment_amount = 0

        self.close_outs = 0

    def initialise(self, vega: VegaServiceNull, create_key: bool = True, mint_key=True):
        super().initialise(vega, create_key)

        self.market_id = self.vega.find_market_id(name=self.market_name)

        # Get asset id
        self.asset_id = self.vega.find_asset_id(symbol=self.asset_name)
        if mint_key:
            # Top up asset
            self.vega.mint(
                key_name=self.key_name,
                asset=self.asset_id,
                amount=self.initial_asset_mint,
                wallet_name=self.wallet_name,
            )

        self.vega.wait_fn(5)

    def step(self, vega_state):
        account = self.vega.party_account(
            key_name=self.key_name,
            wallet_name=self.wallet_name,
            market_id=self.market_id,
            asset_id=self.asset_id,
        )

        total_balance = account.general + account.margin + account.bond

        if total_balance == 0:
            self.vega.mint(
                key_name=self.key_name,
                wallet_name=self.wallet_name,
                amount=self.initial_asset_mint,
                asset=self.asset_id,
            )
            self.close_outs += 1
            self.commitment_amount = 0
            return

        if self.random_state.rand() > self.step_bias:
            return

        if self.commitment_amount < self.commitment_factor * (total_balance):
            self.commitment_amount = self.commitment_factor * (total_balance)

            self.vega.submit_simple_liquidity(
                key_name=self.key_name,
                wallet_name=self.wallet_name,
                market_id=self.market_id,
                fee=0.0001,
                commitment_amount=self.commitment_amount,
                reference_buy="PEGGED_REFERENCE_BEST_BID",
                reference_sell="PEGGED_REFERENCE_BEST_ASK",
                delta_buy=0,
                delta_sell=0,
            )


class FuzzyLiquidityProvider(StateAgentWithWallet):
    NAME_BASE = "fuzzy_liquidity_provider"

    def __init__(
        self,
        key_name: str,
        market_name: str,
        asset_name: str,
        tag: Optional[str] = None,
        wallet_name: Optional[str] = None,
        state_update_freq: Optional[int] = None,
        random_state: Optional[RandomState] = None,
        commitment_factor_min: float = 0.1,
        commitment_factor_max: float = 0.6,
        initial_asset_mint: float = 1e5,
    ):
        super().__init__(key_name, tag, wallet_name, state_update_freq)

        self.market_name = market_name
        self.asset_name = asset_name

        self.commitment_factor_min = commitment_factor_min
        self.commitment_factor_max = commitment_factor_max
        self.random_state = random_state if random_state is not None else RandomState()
        self.initial_asset_mint = initial_asset_mint

    def initialise(self, vega: VegaServiceNull, create_key: bool = True, mint_key=True):
        super().initialise(vega, create_key)

        self.market_id = self.vega.find_market_id(name=self.market_name)

        # Get asset id
        self.asset_id = self.vega.find_asset_id(symbol=self.asset_name)
        if mint_key:
            # Top up asset
            self.vega.mint(
                key_name=self.key_name,
                asset=self.asset_id,
                amount=self.initial_asset_mint,
                wallet_name=self.wallet_name,
            )

        self.vega.wait_fn(5)

    def _gen_spec(self, side: vega_protos.vega.Side, is_valid: bool):
        refs = [
            vega_protos.vega.PeggedReference.PEGGED_REFERENCE_UNSPECIFIED,
            vega_protos.vega.PeggedReference.PEGGED_REFERENCE_MID,
            vega_protos.vega.PeggedReference.PEGGED_REFERENCE_BEST_BID,
            vega_protos.vega.PeggedReference.PEGGED_REFERENCE_BEST_ASK,
        ]
        invalid_ref = (
            vega_protos.vega.PeggedReference.PEGGED_REFERENCE_BEST_BID
            if side == vega_protos.vega.SIDE_SELL
            else vega_protos.vega.PeggedReference.PEGGED_REFERENCE_BEST_ASK
        )
        if is_valid:
            refs = [r for r in refs if r != invalid_ref]

        return (
            self.random_state.choice(refs),
            (
                self.random_state.randint(low=1, high=400)
                if is_valid
                else self.random_state.randint(-5, 5)
            ),
            self.random_state.randint(low=0, high=4000000),
        )

    def step(self, vega_state):
        commitment_factor = (
            self.random_state.random_sample()
            * (self.commitment_factor_max - self.commitment_factor_min)
            + self.commitment_factor_min
        )
        account = self.vega.party_account(
            key_name=self.key_name,
            wallet_name=self.wallet_name,
            market_id=self.market_id,
            asset_id=self.asset_id,
        )

        total_balance = account.general + account.margin + account.bond

        if total_balance == 0:
            self.vega.mint(
                key_name=self.key_name,
                wallet_name=self.wallet_name,
                amount=self.initial_asset_mint,
                asset=self.asset_id,
            )
            return

        commitment_amount = commitment_factor * (total_balance)

        valid = self.random_state.choice([1, 0], p=[0.95, 0.05])
        fee = (
            self.random_state.random_sample() * 0.05
            if valid
            else self.random_state.random_sample() * 200 - 100
        )

        buy_specs = [
            self._gen_spec(vega_protos.vega.SIDE_BUY, valid)
            for _ in range(self.random_state.randint(1, 50))
        ]

        sell_specs = [
            self._gen_spec(vega_protos.vega.SIDE_SELL, valid)
            for _ in range(self.random_state.randint(1, 50))
        ]

        self.vega.submit_liquidity(
            key_name=self.key_name,
            wallet_name=self.wallet_name,
            market_id=self.market_id,
            fee=fee,
            commitment_amount=commitment_amount,
            buy_specs=buy_specs,
            sell_specs=sell_specs,
            is_amendment=self.random_state.choice([True, False]),
        )


class SuccessorMarketCreatorAgent(StateAgentWithWallet):
    NAME_BASE = "successor_market_creator"

    def __init__(
        self,
        key_name: str,
        market_name: str,
        asset_name: str,
        tag: Optional[str] = None,
        wallet_name: Optional[str] = None,
        state_update_freq: Optional[int] = None,
        random_state: Optional[RandomState] = None,
        initial_asset_mint: float = 1e5,
        new_market_prob: float = 0.02,
    ):
        super().__init__(key_name, tag, wallet_name, state_update_freq)

        self.market_name = market_name
        self.asset_name = asset_name

        self.random_state = random_state if random_state is not None else RandomState()
        self.initial_asset_mint = initial_asset_mint

    def initialise(self, vega: VegaServiceNull, create_key: bool = True, mint_key=True):
        super().initialise(vega, create_key)

        self.market_id = self.vega.find_market_id(name=self.market_name)

        # Get asset id
        self.asset_id = self.vega.find_asset_id(symbol=self.asset_name)
        if mint_key:
            # Top up asset
            self.vega.mint(
                key_name=self.key_name,
                asset=self.asset_id,
                amount=self.initial_asset_mint,
                wallet_name=self.wallet_name,
            )

        self.vega.wait_fn(5)


class FuzzySuccessorConfigurableMarketManager(StateAgentWithWallet):
    def __init__(
        self,
        proposal_key_name: str,
        termination_key_name: str,
        market_name: str,
        market_code: str,
        asset_name: str,
        asset_dp: int,
        proposal_wallet_name: Optional[str] = None,
        termination_wallet_name: Optional[str] = None,
        market_config: Optional[MarketConfig] = None,
        tag: Optional[str] = None,
        settlement_price: Optional[float] = None,
        initial_mint: Optional[float] = 1e9,
        successor_probability: float = 0.02,
        random_state: Optional[RandomState] = None,
        market_agents: Optional[Dict[str, List[StateAgentWithWallet]]] = None,
        stake_key: bool = False,
    ):
        super().__init__(
            wallet_name=proposal_wallet_name,
            key_name=proposal_key_name,
            tag=tag,
        )

        self.termination_wallet_name = termination_wallet_name
        self.termination_key_base = termination_key_name
        self.latest_key_idx = 0
        self.market_agents = market_agents if market_agents is not None else {}

        self.successor_probability = successor_probability
        self.random_state = random_state if random_state is not None else RandomState()

        self.market_name_base = market_name
        self.market_code_base = market_code

        self.asset_dp = asset_dp
        self.asset_name = asset_name

        self.initial_mint = initial_mint

        self.market_config = (
            market_config if market_config is not None else MarketConfig()
        )

        self.settlement_price = settlement_price
        self.needs_to_update_markets = False
        self.stake_key = stake_key

    def _get_termination_key_name(self):
        return (
            f"{self.termination_key_base}_{self.latest_key_idx}"
            if self.latest_key_idx > 0
            else self.termination_key_base
        )

    def _get_market_name(self):
        return (
            f"{self.market_name_base}_{self.latest_key_idx}"
            if self.latest_key_idx > 0
            else self.market_name_base
        )

    def _get_market_code(self):
        return (
            f"{self.market_code_base}_{self.latest_key_idx}"
            if self.latest_key_idx > 0
            else self.market_code_base
        )

    def initialise(
        self,
        vega: VegaServiceNull,
        create_key: bool = True,
        mint_key: bool = True,
    ):
        super().initialise(vega=vega, create_key=create_key)
        if create_key:
            self.vega.create_key(
                wallet_name=self.termination_wallet_name,
                name=self._get_termination_key_name(),
            )

        self.vega.wait_for_total_catchup()
        if mint_key:
            self.vega.mint(
                wallet_name=self.wallet_name,
                asset="VOTE",
                amount=1e4,
                key_name=self.key_name,
            )

        if self.stake_key:
            self.vega.stake(
                amount=1,
                key_name=self.key_name,
                wallet_name=self.wallet_name,
            )

        self.vega.wait_for_total_catchup()

        if self.vega.find_asset_id(symbol=self.asset_name) is None:
            self.vega.create_asset(
                wallet_name=self.wallet_name,
                name=self.asset_name,
                symbol=self.asset_name,
                decimals=self.asset_dp,
                quantum=int(10 ** (self.asset_dp)),
                max_faucet_amount=10_000_000_000,
                key_name=self.key_name,
            )

        self.vega.wait_for_total_catchup()
        self.asset_id = self.vega.find_asset_id(symbol=self.asset_name)

        self.vega.wait_for_total_catchup()
        if mint_key:
            self.vega.mint(
                wallet_name=self.wallet_name,
                asset=self.asset_id,
                amount=self.initial_mint,
                key_name=self.key_name,
            )

        self.vega.wait_for_total_catchup()
        self.market_id = self.vega.find_market_id(name=self._get_market_name())

        if self.market_id is None:
            self._create_latest_market()

    def _create_latest_market(self, parent_market_id: Optional[str] = None):
        # Add market information and asset information to market config
        mkt_config = copy.deepcopy(self.market_config)
        mkt_config.set("instrument.name", self._get_market_name())
        mkt_config.set("instrument.code", self._get_market_code())
        mkt_config.set("instrument.future.settlement_asset", self.asset_id)
        mkt_config.set("instrument.future.quote_name", self.asset_name)
        mkt_config.set("instrument.future.number_decimal_places", self.asset_dp)
        mkt_config.set(
            "instrument.future.terminating_key",
            self.vega.wallet.public_key(
                wallet_name=self.termination_wallet_name,
                name=self._get_termination_key_name(),
            ),
        )

        if parent_market_id is not None:
            mkt_config.set(
                "successor",
                Successor(
                    opt={
                        "parent_market_id": parent_market_id,
                        "insurance_pool_fraction": 1,
                    }
                ),
            )

        self.vega.wait_for_total_catchup()
        self.vega.create_market_from_config(
            proposal_wallet_name=self.wallet_name,
            proposal_key_name=self.key_name,
            market_config=mkt_config,
        )

        self.vega.wait_for_total_catchup()
        self.market_id = self.vega.find_market_id(name=self._get_market_name())

    def finalise(self):
        if self.settlement_price is not None:
            self.vega.settle_market(
                self._get_termination_key_name(),
                self.settlement_price,
                self.market_id,
                self.termination_wallet_name,
            )
            self.vega.wait_for_total_catchup()

    def step(self, vega_state) -> None:
        if self.needs_to_update_markets:
            self.vega.settle_market(
                self.old_termination_key,
                self.vega.market_data_from_feed(self.old_market_id).last_traded_price,
                self.market_id,
                self.termination_wallet_name,
            )

            self.market_id = self.vega.find_market_id(name=self._get_market_name())
            for agents in self.market_agents.values():
                for agent in agents:
                    agent.market_id = self.market_id
                    agent.market_name = self._get_market_name()
            self.needs_to_update_markets = False

        if self.random_state.random() < self.successor_probability:
            self.old_market_id = self.market_id
            self.old_termination_key = self._get_termination_key_name()
            self.latest_key_idx += 1

            self.vega.create_key(
                wallet_name=self.termination_wallet_name,
                name=self._get_termination_key_name(),
            )

            self._create_latest_market(parent_market_id=self.market_id)
            self.needs_to_update_markets = True



from abc import ABC
from dataclasses import dataclass
from typing import Any, Tuple, Dict, Optional, List, Union

from vega_sim.null_service import VegaServiceNull
from vega_sim.network_service import VegaServiceNetwork


@dataclass
class VegaState:
    network_state: Tuple
    market_state: Dict[str, Any]


class Agent(ABC):
    NAME_BASE = "base_agent"

    def __init__(self, tag: Optional[str] = None):
        self.tag = tag

    def step(self, vega: Union[VegaServiceNull, VegaServiceNetwork]):
        pass

    def initialise(
        self,
        vega: Union[VegaServiceNull, VegaServiceNetwork],
        create_key: bool = False,
        mint_key: bool = False,
    ):
        self.vega = vega

    def finalise(self):
        pass

    def _update_state(self, current_step: int):
        pass

    def name(self) -> str:
        return self.NAME_BASE + (f"_{self.tag}" if self.tag else "")

    @classmethod
    def name_from_tag(cls, tag: Optional[str] = None) -> str:
        return cls.NAME_BASE + (f"_{tag}" if tag is not None else "")


class AgentWithWallet(Agent):
    def __init__(
        self,
        key_name: str,
        tag: Optional[str] = None,
        wallet_name: Optional[str] = None,
        state_update_freq: Optional[int] = None,
    ):
        """Agent for use in environments as specified in environment.py.
        To extend, the crucial function to implement is the step function which will
        be called on each timestep in the simulation.

        Additionally, the initialise function can be added to. This function is called
        once before the main simulation and can be used to creat assets, set up markets,
        faucet assets to the agent etc.

        Args:
            key_name:
                str, Name of key in wallet for agent to use.
            wallet_pass:
                str, The password which this agent uses to log in to the wallet
            tag:
                str, optional, additional tag to add to agent's wallet name
            wallet_name:
                str, optional, The name to use for this agent's wallet
        """
        super().__init__(tag=tag)
        self.wallet_name = wallet_name
        self.key_name = key_name

        self.state_update_freq = state_update_freq

    def step(self):
        pass

    def initialise(
        self, vega: Union[VegaServiceNull, VegaServiceNetwork], create_key: bool = True
    ):
        super().initialise(vega=vega)
        if create_key:
            self.vega.create_key(
                wallet_name=self.wallet_name,
                name=self.key_name,
            )
            self._public_key = vega.wallet.public_key(
                name=self.key_name, wallet_name=self.wallet_name
            )


class StateAgentWithWallet(AgentWithWallet):
    def step(self, vega_state: VegaState):
        super().step()


class StateAgent(Agent):
    def step(self, vega_state: VegaState):
        super().step()



from __future__ import annotations

import datetime
import logging
from collections import namedtuple
from functools import wraps
from typing import Callable, Iterable, List, Optional, TypeVar, Union, Tuple

import vega_sim.grpc.client as vac
import vega_sim.proto.data_node.api.v2 as data_node_protos_v2
import vega_sim.proto.vega as vega_protos
import vega_sim.proto.vega.events.v1.events_pb2 as events_protos

logger = logging.getLogger(__name__)

MarketAccount = namedtuple("MarketAccount", ["insurance", "liquidity_fee"])

T = TypeVar("T")
S = TypeVar("S")
U = TypeVar("U")


def _retry(num_retry_attempts: int = 3):
    """Automatically retries a function a certain number
    of times. Swallows any errors raised by earlier attempts
    and will raise the last one if still failing.
    """

    def retry_decorator(fn):
        @wraps(fn)
        def auto_retry_fn(*args, **kwargs):
            for i in range(num_retry_attempts):
                try:
                    return fn(*args, **kwargs)
                except Exception as e:
                    if (i + 1) == num_retry_attempts:
                        raise e

        return auto_retry_fn

    return retry_decorator


def unroll_v2_pagination(
    base_request: S,
    request_func: Callable[[S], T],
    extraction_func: Callable[[S], List[U]],
) -> List[T]:
    base_request.pagination.CopyFrom(
        data_node_protos_v2.trading_data.Pagination(first=1000)
    )

    response = request_func(base_request)
    full_list = extraction_func(response)
    while response.page_info.has_next_page:
        base_request.pagination.after = response.page_info.end_cursor
        response = request_func(base_request)
        curr_list = extraction_func(response)
        full_list.extend(curr_list)
    return full_list


@_retry(3)
def positions_by_market(
    pub_key: str,
    data_client: vac.VegaTradingDataClientV2,
    market_id: Optional[str] = None,
) -> List[vega_protos.vega.Position]:
    """Output positions of a party."""

    base_request = data_node_protos_v2.trading_data.ListPositionsRequest(
        party_id=pub_key,
    )
    if market_id is not None:
        setattr(base_request, "market_id", market_id)

    return unroll_v2_pagination(
        base_request=base_request,
        request_func=lambda x: data_client.ListPositions(x).positions,
        extraction_func=lambda res: [i.node for i in res.edges],
    )


@_retry(3)
def all_markets(
    data_client: vac.VegaTradingDataClientV2,
) -> List[vega_protos.markets.Market]:
    """
    Output market info.
    """
    return unroll_v2_pagination(
        base_request=data_node_protos_v2.trading_data.ListMarketsRequest(),
        request_func=lambda x: data_client.ListMarkets(x).markets,
        extraction_func=lambda res: [i.node for i in res.edges],
    )


@_retry(3)
def market_info(
    market_id: str,
    data_client: vac.VegaTradingDataClientV2,
) -> vega_protos.markets.Market:
    """
    Output market info.
    """
    return data_client.GetMarket(
        data_node_protos_v2.trading_data.GetMarketRequest(market_id=market_id)
    ).market


@_retry(3)
def list_assets(data_client: vac.VegaTradingDataClientV2):
    return unroll_v2_pagination(
        base_request=data_node_protos_v2.trading_data.ListAssetsRequest(),
        request_func=lambda x: data_client.ListAssets(x).assets,
        extraction_func=lambda res: [i.node for i in res.edges],
    )


@_retry(3)
def asset_info(
    asset_id: str,
    data_client: vac.VegaTradingDataClientV2,
) -> vac.vega.assets.Asset:
    """Returns information on a given asset selected by its ID

    Args:
        asset_id:
            str, The ID of the asset requested
        data_client:
            VegaTradingDataClientV2, an instantiated gRPC data client

    Returns:
        vega.assets.Asset, Object containing data about the requested asset
    """
    return data_client.GetAsset(
        data_node_protos_v2.trading_data.GetAssetRequest(asset_id=asset_id)
    ).asset


@_retry(3)
def list_accounts(
    data_client: vac.VegaTradingDataClientV2,
    asset_id: Optional[str] = None,
    market_id: Optional[str] = None,
    party_id: Optional[str] = None,
) -> List[data_node_protos_v2.trading_data.AccountBalance]:
    """
    Output liquidity fee account/ insurance pool in the market
    """

    account_filter = data_node_protos_v2.trading_data.AccountFilter()
    if party_id is not None:
        account_filter.party_ids.extend([party_id])
    if asset_id is not None:
        account_filter.asset_id = asset_id
    if market_id is not None:
        account_filter.market_ids.extend([market_id])
    return unroll_v2_pagination(
        base_request=data_node_protos_v2.trading_data.ListAccountsRequest(
            filter=account_filter
        ),
        request_func=lambda x: data_client.ListAccounts(x).accounts,
        extraction_func=lambda res: [i.node for i in res.edges],
    )


@_retry(3)
def market_accounts(
    asset_id: str,
    market_id: str,
    data_client: vac.VegaTradingDataClientV2,
) -> MarketAccount:
    """
    Output liquidity fee account/ insurance pool in the market
    """
    accounts = list_accounts(
        asset_id=asset_id, market_id=market_id, data_client=data_client
    )
    account_type_map = {account.type: account for account in accounts}
    return MarketAccount(
        account_type_map[vega_protos.vega.ACCOUNT_TYPE_INSURANCE],
        account_type_map[vega_protos.vega.ACCOUNT_TYPE_FEES_LIQUIDITY],
    )


@_retry(3)
def get_latest_market_data(
    market_id: str,
    data_client: vac.VegaTradingDataClientV2,
) -> vega_protos.vega.MarketData:
    """
    Output market info.
    """
    return data_client.GetLatestMarketData(
        data_node_protos_v2.trading_data.GetLatestMarketDataRequest(market_id=market_id)
    ).market_data


@_retry(3)
def market_data_history(
    market_id: str,
    start: datetime.datetime,
    end: datetime.datetime,
    data_client: vac.VegaTradingDataClientV2,
) -> List[vega_protos.vega.MarketData]:
    """
    Output market data history.
    """

    return unroll_v2_pagination(
        base_request=data_node_protos_v2.trading_data.GetMarketDataHistoryByIDRequest(
            market_id=market_id,
            start_timestamp=int(start.timestamp() * 1e9),
            end_timestamp=int(end.timestamp() * 1e9),
        ),
        request_func=lambda x: data_client.GetMarketDataHistoryByID(x).market_data,
        extraction_func=lambda res: [i.node for i in res.edges],
    )


@_retry(3)
def infrastructure_fee_accounts(
    asset_id: str,
    data_client: vac.VegaTradingDataClientV2,
) -> List[vega_protos.vega.Account]:
    """
    Output infrastructure fee accounts
    """

    account_filter = data_node_protos_v2.trading_data.AccountFilter(
        asset_id=asset_id,
        account_types=[
            vega_protos.vega.AccountType.ACCOUNT_TYPE_FEES_INFRASTRUCTURE,
        ],
    )

    return unroll_v2_pagination(
        base_request=data_node_protos_v2.trading_data.ListAccountsRequest(
            filter=account_filter
        ),
        request_func=lambda x: data_client.ListAccounts(x).accounts,
        extraction_func=lambda res: [i.node for i in res.edges],
    )


@_retry(3)
def list_orders(
    data_client: vac.VegaTradingDataClientV2,
    market_id: str = None,
    party_id: str = None,
    reference: str = None,
    live_only: bool = True,
) -> List[vega_protos.vega.Orders]:
    """Gets a list of Orders for the specified market and party.

    Function queries the datanode for a response of orders for the specified market_id
    and party_id and state (defaults to live orders only). The function then unrolls
    pagination and returns a list of Orders.

    Args:
        data_client (vac.VegaTradingDataClientV2):
            An instantiated gRPC trading_data_client_V2.
        market_id (str):
            Id for market to return orders from.
        party_id (str):
            Id for party to return orders from.
        live_only (bool, optional):
            Whether to only return live orders. Defaults to True.

    Returns:
        _type_: _description_
    """

    order_filter = data_node_protos_v2.trading_data.OrderFilter(live_only=live_only)

    if reference is not None:
        order_filter.reference = reference

    for attr, val in [
        ("market_ids", [market_id] if market_id is not None else None),
        ("party_ids", [party_id] if party_id is not None else None),
    ]:
        if val is not None:
            getattr(order_filter, attr).extend(val)

    request = data_node_protos_v2.trading_data.ListOrdersRequest(filter=order_filter)

    return unroll_v2_pagination(
        base_request=request,
        request_func=lambda x: data_client.ListOrders(x).orders,
        extraction_func=lambda res: [i.node for i in res.edges],
    )


@_retry(3)
def order_status(
    order_id: str, data_client: vac.VegaTradingDataClientV2, version: int = 0
) -> Optional[vega_protos.vega.Order]:
    """Loads information about a specific order identified by the ID.
    Optionally return historic order versions.

    Args:
        order_id:
            str, the order identifier as specified by Vega when originally placed
        data_client:
            VegaTradingDataClientV2, an instantiated gRPC trading data client
        version:
            int, Optional, Version of the order:
                - Set `version` to 0 for most recent version of the order
                - Set `1` for original version of the order
                - Set `2` for first amendment, `3` for second amendment, etc
    Returns:
        Optional[vega.Order], the requested Order object or None if nothing found
    """
    return data_client.GetOrder(
        data_node_protos_v2.trading_data.GetOrderRequest(
            order_id=order_id, version=version
        )
    ).order


@_retry(3)
def market_depth(
    market_id: str,
    data_client: vac.VegaTradingDataClientV2,
    max_depth: Optional[int] = None,
) -> Optional[vega_protos.vega.MarketDepth]:
    return data_client.GetLatestMarketDepth(
        data_node_protos_v2.trading_data.GetLatestMarketDepthRequest(
            market_id=market_id, max_depth=max_depth
        )
    )


@_retry(3)
def liquidity_provisions(
    data_client: vac.VegaTradingDataClientV2,
    market_id: Optional[str] = None,
    party_id: Optional[str] = None,
) -> Optional[List[vega_protos.vega.LiquidityProvision]]:
    """Loads the current liquidity provision(s) for a given market and/or party.

    Args:
        data_client:
            VegaTradingDataClientV2, an instantiated gRPC trading data client
        market_id:
            Optional[str], the ID of the market from which to pull liquidity provisions
        party_id:
            Optional[str], the ID of the party from which to pull liquidity provisions

    Returns:
        List[LiquidityProvision], list of liquidity provisions (if any exist)
    """
    return unroll_v2_pagination(
        base_request=data_node_protos_v2.trading_data.ListLiquidityProvisionsRequest(
            market_id=market_id, party_id=party_id, live=True
        ),
        request_func=lambda x: data_client.ListLiquidityProvisions(
            x
        ).liquidity_provisions,
        extraction_func=lambda res: [i.node for i in res.edges],
    )


def observe_event_bus(
    data_client: vac.VegaCoreClient,
    type: Optional[list],
    market_id: Optional[str] = None,
    party_id: Optional[str] = None,
) -> Iterable[vega_protos.api.v1.core.ObserveEventBusResponse]:
    """Subscribe to a stream of event updates from the data-node.
    The stream of events returned from this function is an iterable which
    does not end and will continue to tick another order update whenever
    one is received.

    Args:
        type:
            Optional[list], If provided only return events of these types
        market_id:
            Optional[str], If provided, only update events from this market
        party_id:
            Optional[str], If provided, only update events from this party
    Returns:
        Iterable[vega_protos.api.v1.core.ObserveEventBusResponse]:
            Infinite iterable of lists of events updates
    """

    request = vega_protos.api.v1.core.ObserveEventBusRequest(type=type)
    if market_id is not None:
        setattr(request, "market_id", market_id)
    if party_id is not None:
        setattr(request, "party_id", party_id)

    return data_client.ObserveEventBus(iter([request]))


@_retry(3)
def margin_levels(
    data_client: vac.VegaTradingDataClientV2,
    party_id: str,
    market_id: Optional[str] = None,
) -> List[vega_protos.vega.MarginLevels]:
    return unroll_v2_pagination(
        data_node_protos_v2.trading_data.ListMarginLevelsRequest(
            market_id=market_id, party_id=party_id
        ),
        request_func=lambda x: data_client.ListMarginLevels(x).margin_levels,
        extraction_func=lambda res: [i.node for i in res.edges],
    )


@_retry(3)
def get_trades(
    data_client: vac.VegaTradingDataClientV2,
    market_id: str,
    party_id: Optional[str] = None,
    order_id: Optional[str] = None,
    start: Optional[datetime.datetime] = None,
    end: Optional[datetime.datetime] = None,
) -> List[vega_protos.vega.Trade]:
    return unroll_v2_pagination(
        data_node_protos_v2.trading_data.ListTradesRequest(
            market_ids=[market_id],
            party_ids=[party_id] if party_id is not None else None,
            order_ids=[order_id] if order_id is not None else None,
            date_range=data_node_protos_v2.trading_data.DateRange(
                start_timestamp=int(start.timestamp() * 1e9)
                if start is not None
                else None,
                end_timestamp=int(end.timestamp() * 1e9) if end is not None else None,
            ),
        ),
        request_func=lambda x: data_client.ListTrades(x).trades,
        extraction_func=lambda res: [i.node for i in res.edges],
    )


@_retry(3)
def get_network_parameter(
    data_client: vac.VegaTradingDataClientV2,
    key: str,
) -> Union[str, int, float]:
    """Returns the value of the specified network parameter.

    Args:
        data_client (vac.VegaTradingDataClientV2):
            Instantiated gRPC client
        key (str):
            The key identifying the network parameter.

    Returns:
        Any:
            The value of the specified network parameter.
    """
    return data_client.GetNetworkParameter(
        data_node_protos_v2.trading_data.GetNetworkParameterRequest(key=key),
    ).network_parameter


@_retry(3)
def list_transfers(
    data_client: vac.VegaTradingDataClientV2,
    party_id: Optional[str] = None,
    direction: Optional[data_node_protos_v2.trading_data.TransferDirection] = None,
) -> List[events_protos.Transfer]:
    """Returns a list of raw transfers.

    Args:
        data_client (vac.VegaTradingDataClientV2):
            An instantiated gRPC trading data client
        party_id (Optional[str], optional):
            Public key for the specified party. Defaults to None.
        direction (Optional[data_node_protos_v2.trading_data.TransferDirection], optional):
            Direction of transfers to return. Defaults to None.

    Returns:
        List[events_protos.Transfer]:
            A list of Vega Transfer proto objects.
    """

    base_request = data_node_protos_v2.trading_data.ListTransfersRequest(
        pubkey=party_id, direction=direction
    )

    if party_id is not None:
        setattr(base_request, "pubkey", party_id)
    if direction is not None:
        setattr(base_request, "direction", direction)
    else:
        setattr(
            base_request,
            "direction",
            data_node_protos_v2.trading_data.TRANSFER_DIRECTION_TRANSFER_TO_OR_FROM,
        )

    return unroll_v2_pagination(
        base_request=base_request,
        request_func=lambda x: data_client.ListTransfers(x).transfers,
        extraction_func=lambda res: [i.node for i in res.edges],
    )


@_retry(3)
def list_ledger_entries(
    data_client: vac.VegaTradingDataClientV2,
    close_on_account_filters: bool = False,
    asset_id: Optional[str] = None,
    from_party_ids: Optional[List[str]] = None,
    from_market_ids: Optional[List[str]] = None,
    from_account_types: Optional[List[vega_protos.vega.AccountType]] = None,
    to_party_ids: Optional[List[str]] = None,
    to_market_ids: Optional[List[str]] = None,
    to_account_types: Optional[List[vega_protos.vega.AccountType]] = None,
    transfer_types: Optional[List[vega_protos.vega.TransferType]] = None,
    from_datetime: Optional[datetime.datetime] = None,
    to_datetime: Optional[datetime.datetime] = None,
) -> List[data_node_protos_v2.trading_data.AggregatedLedgerEntry]:
    """Returns a list of ledger entries matching specific filters as provided.
    These detail every transfer of funds between accounts within the Vega system,
    including fee/rewards payments and transfers between user margin/general/bond
    accounts.

    Note: At least one of the from_*/to_* filters, or asset ID, must be specified.

    Args:
        data_client:
            vac.VegaTradingDataClientV2, An instantiated gRPC trading data client
        close_on_account_filters:
            bool, default False, Whether both 'from' and 'to' filters must both match
                a given transfer for inclusion. If False, entries matching either
                'from' or 'to' will also be included.
        asset_id:
            Optional[str], filter to only transfers of specific asset ID
        from_party_ids:
            Optional[List[str]], Only include transfers from specified parties
        from_market_ids:
            Optional[List[str]], Only include transfers from specified markets
        from_account_types:
            Optional[List[str]], Only include transfers from specified account types
        to_party_ids:
            Optional[List[str]], Only include transfers to specified parties
        to_market_ids:
            Optional[List[str]], Only include transfers to specified markets
        to_account_types:
            Optional[List[str]], Only include transfers to specified account types
        transfer_types:
            Optional[List[vega_protos.vega.AccountType]], Only include transfers
                of specified types
        from_datetime:
            Optional[datetime.datetime], Only include transfers occurring after
                this time
        to_datetime:
            Optional[datetime.datetime], Only include transfers occurring before
                this time
    Returns:
        List[data_node_protos_v2.trading_data.AggregatedLedgerEntry]
            A list of all transfers matching the requested criteria
    """
    if all(
        not x
        for x in [
            from_party_ids,
            to_party_ids,
            from_account_types,
            to_account_types,
            asset_id,
        ]
    ):
        raise Exception("Must specify at least one filter criterion")

    base_request = data_node_protos_v2.trading_data.ListLedgerEntriesRequest(
        filter=data_node_protos_v2.trading_data.LedgerEntryFilter(
            close_on_account_filters=close_on_account_filters,
            from_account_filter=data_node_protos_v2.trading_data.AccountFilter(
                asset_id=asset_id,
                party_ids=from_party_ids if from_party_ids is not None else [],
                market_ids=from_market_ids if from_market_ids is not None else [],
                account_types=(
                    from_account_types if from_account_types is not None else []
                ),
            ),
            to_account_filter=data_node_protos_v2.trading_data.AccountFilter(
                asset_id=asset_id,
                party_ids=to_party_ids if to_party_ids is not None else [],
                market_ids=to_market_ids if to_market_ids is not None else [],
                account_types=to_account_types if to_account_types is not None else [],
            ),
            transfer_types=transfer_types,
        ),
    )
    if from_datetime is not None or to_datetime is not None:
        base_request.date_range.CopyFrom(
            data_node_protos_v2.trading_data.DateRange(
                start_timestamp=(
                    from_datetime.timestamp() * 1e9
                    if from_datetime is not None
                    else None
                ),
                end_timestamp=(
                    to_datetime.timestamp() * 1e9 if to_datetime is not None else None
                ),
            )
        )

    return unroll_v2_pagination(
        base_request=base_request,
        request_func=lambda x: data_client.ListLedgerEntries(x).ledger_entries,
        extraction_func=lambda res: [i.node for i in res.edges],
    )


def market_data_subscription(
    data_client: vac.VegaCoreClient,
    market_id: Optional[str] = None,
) -> Iterable[vega_protos.vega.MarketData]:
    data_stream = observe_event_bus(
        data_client=data_client,
        type=[events_protos.BUS_EVENT_TYPE_MARKET_DATA],
        market_id=market_id,
    )

    def _data_gen(
        data_stream: Iterable[vega_protos.api.v1.core.ObserveEventBusResponse],
    ) -> Iterable[vega_protos.vega.MarketData]:
        try:
            for market_data in data_stream:
                for market_data_event in market_data.events:
                    yield market_data_event.market_data
        except Exception:
            logger.info("Order subscription closed")
            return

    return _data_gen(data_stream=data_stream)


@_retry(3)
def get_risk_factors(
    data_client: vac.VegaTradingDataClientV2,
    market_id: str,
):
    return data_client.GetRiskFactors(
        data_node_protos_v2.trading_data.GetRiskFactorsRequest(market_id=market_id)
    )


@_retry(3)
def estimate_position(
    data_client: vac.VegaTradingDataClientV2,
    market_id: str,
    open_volume: int,
    orders: Optional[List[data_node_protos_v2.trading_data.OrderInfo]] = None,
    collateral_available: Optional[str] = None,
) -> Tuple[
    data_node_protos_v2.trading_data.MarginEstimate,
    data_node_protos_v2.trading_data.LiquidationEstimate,
]:
    base_request = data_node_protos_v2.trading_data.EstimatePositionRequest(
        market_id=market_id,
        open_volume=open_volume,
    )

    if orders is not None:
        [base_request.orders.append(order) for order in orders]
    if collateral_available is not None:
        setattr(base_request, "collateral_available", collateral_available)

    response = data_client.EstimatePosition(base_request)

    return response.margin, response.liquidation


def get_stake(data_client: vac.trading_data_grpc_v2, party_id: str = None):
    base_request = data_node_protos_v2.trading_data.GetStakeRequest(party_id=party_id)
    return data_client.GetStake(base_request).current_stake_available


def list_all_network_history_segments(data_client: vac.trading_data_grpc_v2):
    base_request = (
        data_node_protos_v2.trading_data.ListAllNetworkHistorySegmentsRequest()
    )
    return data_client.ListAllNetworkHistorySegments(base_request).segments



import logging
import argparse

from typing import Optional

from vega_sim.scenario.constants import Network
from vega_sim.network_service import VegaServiceNetwork
from vega_sim.scenario.fuzzed_markets.scenario import FuzzingScenario


class NetworkHistoryDivergenceError(Exception):
    def __init__(self, from_height, to_height):
        message = f"Divergence occurred between blocks {from_height} and {to_height}"
        super().__init__(message)


def check_divergence(vega):
    # Check all data nodes have matching network history segments
    nodes_checked = set()
    previous_node_segments = None
    while vega.data_node_grpc_url not in nodes_checked:
        # Check the network history segments of the current node match the previous
        current_node_segments = vega.list_all_network_history_segments()
        if previous_node_segments is not None:
            if previous_node_segments != current_node_segments:
                from_height, to_height = find_block_range(
                    current_node_segments, previous_node_segments
                )
                raise NetworkHistoryDivergenceError(from_height, to_height)
        previous_node_segments = current_node_segments

        # Add node to the checked nodes then switch node
        nodes_checked.add(vega.data_node_grpc_url)
        vega.switch_datanode()
    logging.info("Network history segments have not diverged.")


def find_block_range(current_node_segments, previous_node_segments):
    for i in range(1, len(current_node_segments) + 1):
        if (
            current_node_segments[-i].history_segment_id
            != previous_node_segments[-i].history_segment_id
        ):
            return (
                current_node_segments[-i].from_height,
                current_node_segments[-i].to_height,
            )


def _run(
    steps: int,
    test_dir: Optional[str] = None,
    console: Optional[bool] = False,
    network_on_host: bool = False,
):
    scenario = FuzzingScenario(
        num_steps=steps,
        step_length_seconds=2,
        block_length_seconds=1,
        transactions_per_block=4096,
        n_markets=1,
    )

    with VegaServiceNetwork(
        network=Network.CAPSULE,
        run_with_wallet=False,
        run_with_console=console,
        vega_home_path=f"{test_dir}/vegahome",
        wallet_home_path=f"{test_dir}/vegahome/wallet",
        wallet_path="vega_sim/bin/vegawallet",
        network_config_path=f"{test_dir}/vegahome/wallet",
        wallet_passphrase_path="vega_sim/vegacapsule/passphrase-file",
        wallet_url="http://localhost:1789",
        faucet_url="http://localhost:1790",
        vega_node_grpc_url="localhost:3022",
        load_existing_keys=False,
        governance_symbol="VEGA",
        network_on_host=network_on_host,
    ) as vega:
        # Run the fuzzing scenario
        scenario.run_iteration(
            vega=vega,
            network=Network.CAPSULE,
            log_every_n_steps=100,
        )

        check_divergence(vega)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    parser.add_argument("-s", "--steps", type=int, default=100)
    parser.add_argument("--test-dir", type=str, default=None)

    parser.add_argument("-d", "--debug", action="store_true")
    parser.add_argument("--console", action="store_true")
    parser.add_argument("--network_on_host", action="store_true")

    args = parser.parse_args()
    logging.basicConfig(level=logging.DEBUG if args.debug else logging.INFO)

    _run(
        steps=args.steps,
        console=args.console,
        test_dir=args.test_dir,
        network_on_host=args.network_on_host,
    )



import abc
import numpy as np
from typing import Optional, List, Callable, Any, Dict

from vega_sim.null_service import VegaService
from vega_sim.environment.environment import MarketEnvironment
from vega_sim.scenario.constants import Network
from vega_sim.scenario.common.agents import (
    Snitch,
    StateAgent,
    MarketHistoryData,
    Agent,
    ResourceData,
)
from vega_sim.tools.scenario_output import (
    agents_standard_output,
    resources_standard_output,
    market_data_standard_output,
    agents_standard_output,
    assets_standard_output,
    market_chain_standard_output,
)

import vega_sim.proto.vega as vega_protos


class Scenario(abc.ABC):
    def __init__(
        self,
        state_extraction_fn: Optional[
            Callable[[VegaService, Dict[str, Agent]], Any]
        ] = None,
        additional_data_output_fns: Optional[Dict[str, Callable]] = None,
    ):
        self.agents = []
        self.env: Optional[MarketEnvironment] = None
        self.state_extraction_fn = state_extraction_fn
        self.additional_data_output_fns = additional_data_output_fns

    def _step_end_callback(self):
        """Called after each set of agent steps before the next loop begins."""
        pass

    @abc.abstractmethod
    def configure_agents(
        self,
        vega: VegaService,
        tag: str,
        random_state: Optional[np.random.RandomState],
        **kwargs,
    ) -> Dict[str, StateAgent]:
        pass

    @abc.abstractmethod
    def configure_environment(
        self,
        vega: VegaService,
        tag: str,
        random_state: Optional[np.random.RandomState],
        **kwargs,
    ) -> MarketEnvironment:
        pass

    def run(
        self,
        pause_at_completion: bool = False,
        run_with_console: bool = False,
    ):
        result = self.env.run(
            pause_at_completion=pause_at_completion,
            run_with_console=run_with_console,
            step_end_callback=self._step_end_callback,
        )
        return result

    def run_iteration(
        self,
        vega: VegaService,
        pause_at_completion: bool = False,
        run_with_console: bool = False,
        random_state: Optional[np.random.RandomState] = None,
        run_with_snitch: bool = True,
        tag: Optional[str] = None,
        output_data: bool = False,
        log_every_n_steps: Optional[int] = None,
        **kwargs,
    ):
        tag = tag if tag is not None else ""
        self.agents = self.configure_agents(
            vega=vega, tag=tag, random_state=random_state, **kwargs
        )

        if run_with_snitch or output_data:
            self.agents["snitch"] = Snitch(
                agents=self.agents, additional_state_fn=self.state_extraction_fn
            )

        self.env = self.configure_environment(
            vega=vega, tag=tag, random_state=random_state, **kwargs
        )

        outputs = self.env.run(
            pause_at_completion=pause_at_completion,
            run_with_console=run_with_console,
            log_every_n_steps=log_every_n_steps,
            step_end_callback=self._step_end_callback,
        )
        if output_data:
            agents_standard_output(self.agents)
            resources_standard_output(self.get_resource_data())
            assets_standard_output(self.get_assets())
            market_data_standard_output(self.get_run_data())
            market_chain_standard_output(self.get_run_data())
            if self.additional_data_output_fns is not None:
                market_data_standard_output(
                    self.get_additional_run_data(),
                    custom_output_fns=self.additional_data_output_fns,
                )

        return outputs

    def get_snitch(self) -> Optional[Snitch]:
        return self.agents.get("snitch")

    def get_resource_data(self) -> List[ResourceData]:
        snitch = self.get_snitch()
        return snitch.resources if snitch is not None else []

    def get_assets(self) -> List[vega_protos.assets.Asset]:
        snitch = self.get_snitch()
        return snitch.assets if snitch is not None else []

    def get_run_data(self) -> List[MarketHistoryData]:
        snitch = self.get_snitch()
        return snitch.states if snitch is not None else []

    def get_additional_run_data(self) -> List[MarketHistoryData]:
        snitch = self.get_snitch()
        return snitch.additional_states if snitch is not None else []



from abc import abstractmethod
from collections import defaultdict, namedtuple
from typing import Dict, Optional, Tuple


from vega_sim.environment import VegaState
from vega_sim.environment.agent import Agent
from vega_sim.null_service import VegaServiceNull


from vega_sim.reinforcement.v2.states import State
from vega_sim.reinforcement.v2.agents.puppets import Action

WalletConfig = namedtuple("WalletConfig", ["name", "passphrase"])

# Send selling/buying MOs to hit LP orders
WALLET = WalletConfig("learner", "learner")


def state_fn(
    service: VegaServiceNull,
    agents: Dict[str, Agent],
    state_values=None,
) -> Tuple[State, Action]:
    learner = agents["learner"]
    return (learner.latest_state, learner.latest_action)


class LearningAgent:
    @abstractmethod
    def learning_step(self, results_dir: Optional[str] = None):
        pass

    @abstractmethod
    def update_memory(self, state: State, action: Action, reward: float):
        """
        Updates memory of the agent
        """
        pass

    def clear_memory(self):
        pass

    @abstractmethod
    def step(self, state: State) -> Action:
        pass

    @abstractmethod
    def save(self, results_dir: str):
        pass

    @abstractmethod
    def load(self, results_dir: str):
        pass


class TorchLearningAgent(LearningAgent):
    @abstractmethod
    def __init__(
        self,
        device: str,
        logfile_pol_imp: str,
        logfile_pol_eval: str,
        logfile_pnl: str,
        discount_factor: float,
        inventory_penalty: float = 0.0,
    ):
        super().__init__()

        self.step_num = 0
        self.latest_state = None
        self.latest_action = None
        self.device = device
        self.discount_factor = discount_factor

        self.memory = defaultdict(list)
        self.memory_capacity = 100_000

        # Coefficients for regularisation
        self.coefH_discr = 1.0
        self.coefH_cont = 0.01
        # losses logger
        self.losses = defaultdict(list)
        # logfile
        self.logfile_pol_imp = logfile_pol_imp
        self.logfile_pol_eval = logfile_pol_eval
        self.logfile_pnl = logfile_pnl

        self.learningIteration = 0

    @abstractmethod
    def move_to_device(self):
        pass

    @abstractmethod
    def move_to_cpu(self):
        pass

    @abstractmethod
    def learning_step(self, results_dir: Optional[str] = None):
        pass

    def _update_memory(
        self,
        state: State,
        action: Action,
        reward: float,
    ):
        pass

    def update_memory(self, state: State, action: Action, reward: float):
        """
        Updates memory of the agent, and removes old tuples (s,a,r,s) if memory exceeds its capacity
        """
        self._update_memory(state, action, reward)
        # remove old tuples if memory exceeds its capaciy
        for key in self.memory.keys():
            if len(self.memory[key]) > self.memory_capacity:
                first_n = len(self.memory[key]) - self.memory_capacity
                del self.memory[key][:first_n]
        return 0

    def clear_memory(self):
        for key in self.memory.keys():
            self.memory[key].clear()

    @abstractmethod
    def create_dataloader(self, batch_size):
        """
        creates dataset and dataloader for training.
        """
        pass

    @abstractmethod
    def empty_action(self) -> Action:
        pass

    @abstractmethod
    def step(self, vega_state: VegaState):
        pass

    @abstractmethod
    def save(self, results_dir: str):
        pass

    @abstractmethod
    def load(self, results_dir: str):
        pass



"""registry.py

Module contains preconfigured scenarios.

"""

from vega_sim.devops.scenario import DevOpsScenario

from vega_sim.scenario.common.agents import (
    ArbitrageLiquidityProvider,
    ExponentialShapedMarketMaker,
)

from vega_sim.devops.classes import (
    MarketMakerArgs,
    MarketManagerArgs,
    AuctionTraderArgs,
    RandomTraderArgs,
    SensitiveTraderArgs,
    SimulationArgs,
)

from vega_sim.scenario.common.utils.price_process import Granularity, LivePrice

SCENARIOS = {
    "ETHUSD": lambda: DevOpsScenario(
        binance_code="ETHDAI",
        market_manager_args=MarketManagerArgs(
            market_name="ETHUSD",
            market_code="ETHUSD",
            asset_name="tDAI",
            adp=5,
            mdp=5,
            pdp=3,
        ),
        market_maker_args=MarketMakerArgs(
            market_kappa=10,
            market_order_arrival_rate=100,
            order_kappa=0.3,
            order_size=1,
            order_levels=25,
            order_spacing=0.5,
            order_clipping=10000,
            inventory_lower_boundary=-30,
            inventory_upper_boundary=30,
            fee_amount=0.001,
            commitment_amount=1e5,
            initial_mint=2e5,
        ),
        auction_trader_args=AuctionTraderArgs(
            initial_volume=0.001,
            initial_mint=1e4,
        ),
        random_trader_args=RandomTraderArgs(
            order_intensity=[5, 5, 5],
            order_volume=[0.01, 0.1, 1],
            step_bias=[0.333, 0.012, 0.003],
            initial_mint=1e6,
        ),
        sensitive_trader_args=SensitiveTraderArgs(
            scale=[10, 10, 10],
            max_order_size=[0.001, 0.01, 0.1],
            initial_mint=1e4,
        ),
        simulation_args=SimulationArgs(
            n_steps=60 * 6,
            granularity=Granularity.MINUTE,
            coinbase_code="ETH-USD",
            start_date="2022-11-01 00:00:00",
            randomise_history=False,
        ),
    ),
    "BTCUSD": lambda: DevOpsScenario(
        binance_code="BTCDAI",
        market_manager_args=MarketManagerArgs(
            market_name="BTCUSD Monthly (Dec 2022)",
            market_code="BTCUSD",
            asset_name="tDAI",
            adp=5,
            mdp=5,
            pdp=3,
        ),
        market_maker_args=MarketMakerArgs(
            market_kappa=0.15,
            market_order_arrival_rate=100,
            order_kappa=0.15,
            order_size=1,
            order_levels=25,
            order_spacing=1,
            order_clipping=10000,
            inventory_lower_boundary=-3,
            inventory_upper_boundary=3,
            fee_amount=0.0001,
            commitment_amount=1e5,
            initial_mint=2e5,
        ),
        auction_trader_args=AuctionTraderArgs(
            initial_volume=0.001,
            initial_mint=1e4,
        ),
        random_trader_args=RandomTraderArgs(
            order_intensity=[5, 5, 5],
            order_volume=[0.01, 0.1, 1],
            step_bias=[0.333, 0.012, 0.003],
            initial_mint=1e6,
        ),
        sensitive_trader_args=SensitiveTraderArgs(
            scale=[10, 10, 10],
            max_order_size=[0.001, 0.01, 0.1],
            initial_mint=1e4,
        ),
        simulation_args=SimulationArgs(
            n_steps=60 * 6,
            granularity=Granularity.MINUTE,
            coinbase_code="BTC-USDT",
            start_date="2022-11-01 00:00:00",
            randomise_history=False,
        ),
    ),
    "ADAUSDT": lambda: DevOpsScenario(
        binance_code="ADAUSDT",
        market_manager_args=MarketManagerArgs(
            market_name="Cardano USD",
            market_code="ADA/USD",
            asset_name="tDAI",
            adp=18,
            mdp=4,
            pdp=0,
        ),
        market_maker_args=MarketMakerArgs(
            market_kappa=10000,
            market_order_arrival_rate=5000,
            order_kappa=350,
            order_size=1,
            order_levels=25,
            order_spacing=0.0002,
            order_clipping=200000,
            inventory_lower_boundary=-5000,
            inventory_upper_boundary=5000,
            fee_amount=0.0001,
            commitment_amount=3e5,
            initial_mint=15e5,
        ),
        auction_trader_args=AuctionTraderArgs(
            initial_volume=1,
            initial_mint=5e4,
        ),
        random_trader_args=RandomTraderArgs(
            order_intensity=[100, 100, 100],
            order_volume=[1, 1, 1],
            step_bias=[1, 1, 1],
            initial_mint=5e4,
        ),
        sensitive_trader_args=SensitiveTraderArgs(
            scale=[10, 10, 10],
            max_order_size=[0.001, 0.01, 0.1],
            initial_mint=1e4,
        ),
        simulation_args=SimulationArgs(
            n_steps=6 * 60,
            granularity=Granularity.MINUTE,
            coinbase_code="ADA-USDT",
            start_date="2022-11-01 00:00:00",
            randomise_history=False,
        ),
    ),
    "XRPUSDT": lambda: DevOpsScenario(
        binance_code="XRPUSDT",
        market_manager_args=MarketManagerArgs(
            market_name="XRP USD",
            market_code="XRP USD",
            asset_name="tDAI",
            adp=18,
            mdp=4,
            pdp=0,
        ),
        market_maker_args=MarketMakerArgs(
            market_kappa=10000,
            market_order_arrival_rate=5000,
            order_kappa=350,
            order_size=1,
            order_levels=25,
            order_spacing=0.0002,
            order_clipping=200000,
            inventory_lower_boundary=-5000,
            inventory_upper_boundary=5000,
            fee_amount=0.0001,
            commitment_amount=3e5,
            initial_mint=15e5,
        ),
        auction_trader_args=AuctionTraderArgs(
            initial_volume=1,
            initial_mint=5e4,
        ),
        random_trader_args=RandomTraderArgs(
            order_intensity=[100, 100, 100],
            order_volume=[1, 1, 1],
            step_bias=[1, 1, 1],
            initial_mint=5e4,
        ),
        sensitive_trader_args=SensitiveTraderArgs(
            scale=[10, 10, 10],
            max_order_size=[0.001, 0.01, 0.1],
            initial_mint=1e4,
        ),
        simulation_args=SimulationArgs(
            n_steps=6 * 60,
            granularity=Granularity.MINUTE,
            coinbase_code="ADA-USDT",
            start_date="2023-01-01 00:00:00",
            randomise_history=False,
        ),
    ),
}

AGENTS = {
    "arbitrage_liquidity_provider": lambda: ArbitrageLiquidityProvider(
        wallet_name=None,
        key_name=None,
        market_name=None,
        asset_name=None,
        initial_asset_mint=500,
        commitment_ratio=0.5,
        safety_factor=0.1,
        fee=0.001,
        tag="agent",
    ),
    "shaped_market_maker_ethusd": lambda: ExponentialShapedMarketMaker(
        wallet_name=None,
        key_name=None,
        market_name=None,
        asset_name=None,
        initial_asset_mint=1e9,
        commitment_amount=20000,
        market_kappa=10,
        kappa=0.3,
        num_levels=25,
        tick_spacing=0.5,
        max_order_size=10000,
        inventory_lower_boundary=-30,
        inventory_upper_boundary=30,
        market_order_arrival_rate=100,
        fee_amount=0.0005,
        num_steps=60 * 60 * 24 * 365,
        price_process_generator=iter(LivePrice(product="ETHDAI")),
        orders_from_stream=False,
        state_update_freq=10,
        tag="agent",
    ),
}
